# Licensed under the GPLv3 - see LICENSE
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import numpy as np
import pytest
from astropy.time import Time
import astropy.units as u
from astropy.tests.helper import catch_warnings

from ... import vdif, vlbi_base
from ...data import (SAMPLE_VDIF as SAMPLE_FILE, SAMPLE_VLBI_VDIF as
                     SAMPLE_VLBI, SAMPLE_MWA_VDIF as SAMPLE_MWA,
                     SAMPLE_AROCHIME_VDIF as SAMPLE_AROCHIME)

# Comparison with m5access routines (check code on 2015-MAY-30) on vlba.m5a,
# which contains the first 16 frames from evn/Fd/GP052D_FD_No0006.m5a.
# 00000000  77 2c db 00 00 00 00 1c  75 02 00 20 fc ff 01 04  # header 0 - 3
# 00000010  10 00 80 03 ed fe ab ac  00 00 40 33 83 15 03 f2  # header 4 - 7
# 00000020  2a 0a 7c 43 8b 69 9d 59  cb 99 6d 9a 99 96 5d 67  # data 0 - 3
# NOTE: thread_id = 1
# 2a = 00 10 10 10 = (lsb first) 1,  1,  1, -3
# 0a = 00 00 10 10 =             1,  1, -3, -3
# 7c = 01 11 11 00 =            -3,  3,  3, -1
# m5d evn/Fd/GP052D_FD_No0006.m5a VDIF_5000-512-1-2 100
# Mark5 stream: 0x16cd140
#   stream = File-1/1=evn/Fd/GP052D_FD_No0006.m5a
#   format = VDIF_5000-512-1-2 = 3
#   start mjd/sec = 56824 21367.000000000
#   frame duration = 78125.00 ns
#   framenum = 0
#   sample rate = 256000000 Hz
#   offset = 0
#   framebytes = 5032 bytes
#   datasize = 5000 bytes
#   sample granularity = 4
#   frame granularity = 1
#   gframens = 78125
#   payload offset = 32
#   read position = 0
#   data window size = 1048576 bytes


class TestVDIF(object):

    def test_header(self, tmpdir):
        with open(SAMPLE_FILE, 'rb') as fh:
            header = vdif.VDIFHeader.fromfile(fh)
        assert header.nbytes == 32
        assert header.edv == 3
        mjd, frac = divmod(header.time.mjd, 1)
        assert mjd == 56824
        assert round(frac * 86400) == 21367
        assert header.payload_nbytes == 5000
        assert header.frame_nbytes == 5032
        assert header['thread_id'] == 1
        assert header.sample_rate == 32*u.MHz
        assert header.samples_per_frame == 20000
        assert header.nchan == 1
        assert header.bps == 2
        assert not header['complex_data']
        assert header.mutable is False
        with open(str(tmpdir.join('test.vdif')), 'w+b') as s:
            header.tofile(s)
            s.seek(0)
            header2 = vdif.VDIFHeader.fromfile(s)
        assert header2 == header
        assert header2.mutable is False
        header3 = vdif.VDIFHeader.fromkeys(**header)
        assert header3 == header
        assert header3.mutable is True
        with pytest.raises(KeyError):
            vdif.VDIFHeader.fromkeys(extra=1, **header)
        with pytest.raises(KeyError):
            kwargs = dict(header)
            kwargs.pop('thread_id')
            vdif.VDIFHeader.fromkeys(**kwargs)
        # Try initialising with properties instead of keywords, as much as
        # possible.  Note that we still have to give a lots of extra, less
        # directly useful parameters to get an *identical* header.
        header4 = vdif.VDIFHeader.fromvalues(
            edv=header.edv, ref_epoch=header['ref_epoch'],
            seconds=header['seconds'], frame_nr=header['frame_nr'],
            samples_per_frame=header.samples_per_frame,
            bps=header.bps, complex_data=header['complex_data'],
            thread_id=header['thread_id'], station=header.station,
            sampling_unit=header['sampling_unit'],
            sampling_rate=header['sampling_rate'],
            loif_tuning=header['loif_tuning'], dbe_unit=header['dbe_unit'],
            if_nr=header['if_nr'], subband=header['subband'],
            sideband=header['sideband'], major_rev=header['major_rev'],
            minor_rev=header['minor_rev'], personality=header['personality'],
            _7_28_4=header['_7_28_4'])
        # The same header, but created by passing time and sample rate.
        header4_usetime = vdif.VDIFHeader.fromvalues(
            edv=header.edv, time=header.time,
            samples_per_frame=header.samples_per_frame,
            station=header.station, sample_rate=header.sample_rate,
            bps=header.bps, complex_data=header['complex_data'],
            thread_id=header['thread_id'],
            loif_tuning=header['loif_tuning'], dbe_unit=header['dbe_unit'],
            if_nr=header['if_nr'], subband=header['subband'],
            sideband=header['sideband'], major_rev=header['major_rev'],
            minor_rev=header['minor_rev'], personality=header['personality'],
            _7_28_4=header['_7_28_4'])
        assert header4 == header
        assert header4.mutable is True
        assert header4 == header4_usetime
        header5 = header.copy()
        assert header5 == header
        assert header5.mutable is True
        header5['thread_id'] = header['thread_id'] + 1
        assert header5['thread_id'] == header['thread_id'] + 1
        assert header5 != header
        with pytest.raises(TypeError):
            header['thread_id'] = 0
        # Also test time setting.
        header5.time = header.time + 1.*u.s
        frame_rate = header.sample_rate / header.samples_per_frame
        assert abs(header5.time - header.time - 1.*u.s) < 1.*u.ns
        assert header5['frame_nr'] == header['frame_nr']
        header5.time = header.time + 1.*u.s + 1.1/frame_rate
        assert abs(header5.time - header.time - 1.*u.s -
                   1./frame_rate) < 1.*u.ns
        assert header5['frame_nr'] == header['frame_nr'] + 1
        # Check rounding in corner case.
        header5.time = header.time + 1.*u.s - 0.01/frame_rate
        assert abs(header5.time - header.time - 1.*u.s) < 1.*u.ns
        assert header5['frame_nr'] == header['frame_nr']
        # Check requesting non-existent EDV returns VDIFBaseHeader instance
        header6 = vdif.header.VDIFHeader.fromvalues(edv=100)
        assert type(header6) is vdif.header.VDIFBaseHeader
        assert header6['edv'] == 100

        # Make a new header to test passing time/sample rate.
        headerT = header.copy()
        headerT.time = header.time + 1. / frame_rate

        # Test initializing EDV 0 with properties, but off of 1 second mark so
        # frame_nr is used.
        header8 = vdif.VDIFHeader.fromvalues(
            edv=0, ref_epoch=headerT['ref_epoch'], seconds=headerT['seconds'],
            frame_nr=headerT['frame_nr'], complex_data=headerT['complex_data'],
            samples_per_frame=headerT.samples_per_frame, bps=headerT.bps,
            station=headerT.station, thread_id=headerT['thread_id'])
        assert header8['ref_epoch'] == headerT['ref_epoch']
        assert header8['seconds'] == headerT['seconds']
        assert header8['frame_nr'] == headerT['frame_nr']
        # The same header, but created by passing time and sample rate.
        header8_usetime = vdif.VDIFHeader.fromvalues(
            edv=0, time=headerT.time, sample_rate=headerT.sample_rate,
            complex_data=headerT['complex_data'], bps=headerT.bps,
            samples_per_frame=headerT.samples_per_frame,
            station=headerT.station, thread_id=headerT['thread_id'])
        assert header8_usetime == header8

        # Without a sample rate or frame_nr, cannot initialize using time.
        with pytest.raises(ValueError):
            vdif.VDIFHeader.fromvalues(
                edv=0, time=headerT.time, complex_data=headerT['complex_data'],
                bps=headerT.bps, samples_per_frame=headerT.samples_per_frame,
                station=headerT.station, thread_id=headerT['thread_id'])

        # Without a sample rate for EDV 1, 3, cannot initialize using time.
        with pytest.raises(ValueError):
            vdif.VDIFHeader.fromvalues(
                edv=1, time=headerT.time, station=headerT.station,
                samples_per_frame=headerT.samples_per_frame,
                bps=headerT.bps, complex_data=headerT['complex_data'],
                thread_id=headerT['thread_id'])

        # Check rounding in corner case.
        header9 = headerT.copy()
        header9.set_time(Time('2018-01-01T00:34:07.999999999996'))
        assert header9['seconds'] == 2048
        assert header9['frame_nr'] == 0

        # Check that with a missing sampling rate, we do not return a wrong
        # time.
        header10 = headerT.copy()
        header10['sampling_rate'] = 0
        with pytest.raises(ValueError):
            header10.time
        assert (header10.get_time(frame_rate=headerT.frame_rate) ==
                headerT.time)

    def test_custom_header(self, tmpdir):
        # Custom header with an EDV that already exists
        with pytest.raises(ValueError):
            class VDIFHeaderY(vdif.header.VDIFBaseHeader):
                _edv = 3
        # Custom header that has neglected to override _edv
        with pytest.raises(ValueError):
            class VDIFHeaderZ(vdif.header.VDIFBaseHeader):
                pass

        # Working header with nonsense data in the last two words.
        class VDIFHeaderX(vdif.header.VDIFSampleRateHeader):
            _edv = 0x58
            _header_parser = (vdif.header.VDIFSampleRateHeader._header_parser +
                              vlbi_base.header.HeaderParser(
                                  (('nonsense_0', (6, 0, 32, 0x0)),
                                   ('nonsense_1', (7, 0, 8, None)),
                                   ('nonsense_2', (7, 8, 24, 0x1)))))

            def verify(self):
                super(VDIFHeaderX, self).verify()

        assert vdif.header.VDIF_HEADER_CLASSES[0x58] is VDIFHeaderX

        # Read in a header, and hack an 0x58 header with its info
        with open(SAMPLE_FILE, 'rb') as fh:
            header = vdif.VDIFHeader.fromfile(fh)

        headerX_dummy = vdif.VDIFHeader.fromvalues(
            edv=0x58, time=header.time,
            samples_per_frame=header.samples_per_frame,
            station=header.station, sample_rate=header.sample_rate,
            bps=header.bps, complex_data=header['complex_data'],
            thread_id=header['thread_id'], nonsense_0=2000000000,
            nonsense_1=100, nonsense_2=10000000)

        # Check that headerX_dummy is indeed VDIFHeaderX class, with
        # the correct EDV
        assert isinstance(headerX_dummy, VDIFHeaderX)
        assert headerX_dummy.edv == 0x58

        # Write to dummy file, then re-read.
        with open(str(tmpdir.join('test.vdif')), 'w+b') as s:
            headerX_dummy.tofile(s)
            s.seek(0)
            headerX = vdif.VDIFHeader.fromfile(s)

        # Ensure values have been copied and re-read properly - only
        # edv, mutability and nonsense values should have changed or
        # been added.
        assert isinstance(headerX, VDIFHeaderX)
        assert headerX.nbytes == header.nbytes
        assert headerX.edv == 0x58
        mjd, frac = divmod(header.time.mjd, 1)
        assert mjd == 56824
        assert round(frac * 86400) == 21367
        assert headerX.payload_nbytes == header.payload_nbytes
        assert headerX.frame_nbytes == header.frame_nbytes
        assert headerX['thread_id'] == header['thread_id']
        assert headerX.sample_rate == header.sample_rate
        assert headerX.samples_per_frame == header.samples_per_frame
        assert headerX.nchan == header.nchan
        assert headerX.bps == header.bps
        assert not headerX['complex_data']
        assert headerX.mutable is False
        assert headerX['nonsense_0'] == 2000000000
        assert headerX['nonsense_1'] == 100
        assert headerX['nonsense_2'] == 10000000

    def test_decoding(self, tmpdir):
        """Check that look-up levels are consistent with mark5access."""
        o2h = vlbi_base.encoding.OPTIMAL_2BIT_HIGH
        assert np.all(vdif.payload.lut1bit[0] == -1.)
        assert np.all(vdif.payload.lut1bit[0xff] == 1.)
        assert np.all(vdif.payload.lut1bit.astype(int) ==
                      ((np.arange(256)[:, np.newaxis] >>
                        np.arange(8)) & 1) * 2 - 1)
        assert np.all(vdif.payload.lut2bit[0] == -o2h)
        assert np.all(vdif.payload.lut2bit[0x55] == -1.)
        assert np.all(vdif.payload.lut2bit[0xaa] == 1.)
        assert np.all(vdif.payload.lut2bit[0xff] == o2h)
        assert np.allclose(vdif.payload.lut4bit[0] * 2.95, -8.)
        assert np.all(vdif.payload.lut4bit[0x88] == 0.)
        assert np.allclose(vdif.payload.lut4bit[0xff] * 2.95, 7)
        assert np.all(-vdif.payload.lut4bit[0x11] ==
                      vdif.payload.lut4bit[0xff])
        aint = np.arange(0, 256, dtype=np.uint8)
        words = aint.view('<u4')
        areal = np.linspace(-127.5, 127.5, 256).reshape(-1, 1) / 35.5
        acmplx = areal[::2] + 1j * areal[1::2]
        payload1 = vdif.VDIFPayload(words, bps=8, complex_data=False)
        assert np.allclose(payload1.data, areal)
        payload2 = vdif.VDIFPayload(words, bps=8, complex_data=True)
        assert np.allclose(payload2.data, acmplx)
        header = vdif.VDIFHeader.fromvalues(edv=0, complex_data=False, bps=8,
                                            payload_nbytes=payload1.nbytes)
        assert vdif.VDIFPayload.fromdata(areal, header) == payload1
        # Also check that a circular decode-encode is self-consistent.
        assert vdif.VDIFPayload.fromdata(payload1.data, header) == payload1
        header['complex_data'] = True
        assert vdif.VDIFPayload.fromdata(acmplx, header) == payload2
        assert vdif.VDIFPayload.fromdata(payload2.data, header) == payload2
        # Also check for bps=4.
        decode = (aint[:, np.newaxis] >> np.array([0, 4])) & 0xf
        areal = ((decode - 8.) / 2.95).reshape(-1, 1)
        acmplx = areal[::2] + 1j * areal[1::2]
        payload3 = vdif.VDIFPayload(words, bps=4, complex_data=False)
        assert np.allclose(payload3.data, areal)
        payload4 = vdif.VDIFPayload(words, bps=4, complex_data=True)
        assert np.allclose(payload4.data, acmplx)
        header = vdif.VDIFHeader.fromvalues(edv=0, complex_data=False, bps=4,
                                            payload_nbytes=payload3.nbytes)
        assert vdif.VDIFPayload.fromdata(areal, header) == payload3
        assert vdif.VDIFPayload.fromdata(payload3.data, header) == payload3
        header['complex_data'] = True
        assert vdif.VDIFPayload.fromdata(acmplx, header) == payload4
        assert vdif.VDIFPayload.fromdata(payload4.data, header) == payload4

    def test_payload(self, tmpdir):
        with open(SAMPLE_FILE, 'rb') as fh:
            header = vdif.VDIFHeader.fromfile(fh)
            payload = vdif.VDIFPayload.fromfile(fh, header)
        assert payload.nbytes == 5000
        assert payload.shape == (20000, 1)
        assert payload.size == 20000
        assert payload.ndim == 2
        # Check sample shape validity
        assert payload.sample_shape == (1,)
        assert payload.sample_shape.nchan == 1
        assert payload.dtype == np.float32
        assert np.all(payload[:12, 0].astype(int) ==
                      np.array([1, 1, 1, -3, 1, 1, -3, -3, -3, 3, 3, -1]))
        with open(str(tmpdir.join('test.vdif')), 'w+b') as s:
            payload.tofile(s)
            s.seek(0)
            payload2 = vdif.VDIFPayload.fromfile(s, header)
            assert payload2 == payload
            with pytest.raises(EOFError):
                # Too few bytes.
                s.seek(100)
                vdif.VDIFPayload.fromfile(s, header)
        payload3 = vdif.VDIFPayload.fromdata(payload.data, header)
        assert payload3 == payload
        with pytest.raises(ValueError):
            # Wrong number of channels.
            vdif.VDIFPayload.fromdata(np.empty((payload.shape[0], 2)), header)
        with pytest.raises(ValueError):
            # Too few data.
            vdif.VDIFPayload.fromdata(payload[:100], header)
        # check if it works with complex data
        payload4 = vdif.VDIFPayload(payload.words, nchan=1, bps=2,
                                    complex_data=True)
        assert payload4.complex_data is True
        assert payload4.nbytes == 5000
        assert payload4.shape == (10000, 1)
        assert payload4.dtype == np.complex64
        assert np.all(payload4.data ==
                      payload[::2] + 1j * payload[1::2])
        with pytest.raises(ValueError):
            vdif.VDIFPayload.fromdata(payload4.data, header)
        header5 = header.copy()
        header5['complex_data'] = True
        payload5 = vdif.VDIFPayload.fromdata(payload4.data, header5)
        assert payload5 == payload4

    @pytest.mark.parametrize('item', (2, (), -1, slice(1, 3),
                                      slice(2, 4), slice(-3, None)))
    def test_payload_getitem_setitem(self, item):
        with open(SAMPLE_FILE, 'rb') as fh:
            header = vdif.VDIFHeader.fromfile(fh)
            payload = vdif.VDIFPayload.fromfile(fh, header)
        sel_data = payload.data[item]
        assert np.all(payload[item] == sel_data)
        payload2 = vdif.VDIFPayload(payload.words.copy(), header)
        assert payload2 == payload
        payload2[item] = -sel_data
        check = payload.data
        check[item] = -sel_data
        assert np.all(payload2[item] == -sel_data)
        assert np.all(payload2.data == check)
        assert payload2 != payload
        payload2[item] = sel_data
        assert np.all(payload2[item] == sel_data)
        assert payload2 == payload

    def test_filereader(self):
        with vdif.open(SAMPLE_FILE, 'rb') as fh:
            header = vdif.VDIFHeader.fromfile(fh)
            fh.seek(0)
            header2 = fh.read_header()
            assert header2 == header
            current_pos = fh.tell()
            frame_rate = fh.get_frame_rate()
            assert abs(frame_rate - 32. * u.MHz /
                       header.samples_per_frame) < 1. * u.nHz
            assert fh.tell() == current_pos
            # The read_frame method is tested below, as is mode='wb'.

    def test_frame(self, tmpdir):
        with vdif.open(SAMPLE_FILE, 'rb') as fh:
            header = vdif.VDIFHeader.fromfile(fh)
            payload = vdif.VDIFPayload.fromfile(fh, header)
            fh.seek(0)
            frame = fh.read_frame()

        assert frame.header == header
        assert frame.payload == payload
        assert frame.shape == payload.shape
        assert frame.size == payload.size
        assert frame.ndim == payload.ndim
        assert frame == vdif.VDIFFrame(header, payload)
        assert np.all(frame.data[:12, 0].astype(int) ==
                      np.array([1, 1, 1, -3, 1, 1, -3, -3, -3, 3, 3, -1]))
        vdif_test = str(tmpdir.join('test.vdif'))
        with open(vdif_test, 'w+b') as s:
            frame.tofile(s)
            s.seek(0)
            frame2 = vdif.VDIFFrame.fromfile(s)

        assert frame2 == frame
        frame3 = vdif.VDIFFrame.fromdata(payload.data, header)
        assert frame3 == frame
        frame4 = vdif.VDIFFrame.fromdata(payload.data, **header)
        assert frame4 == frame
        header5 = header.copy()
        frame5 = vdif.VDIFFrame(header5, payload, valid=False)
        assert frame5.valid is False
        assert np.all(frame5.data == 0.)
        frame5.valid = True
        assert frame5 == frame
        # Also test binary file writer.
        with vdif.open(vdif_test, 'wb') as fw:
            fw.write_frame(frame)
            fw.write_frame(frame.data, header)
        with open(vdif_test, 'rb') as s:
            frame6 = vdif.VDIFFrame.fromfile(s)
            frame7 = vdif.VDIFFrame.fromfile(s)
        assert frame6 == frame
        assert frame7 == frame

    def test_frameset(self):
        with vdif.open(SAMPLE_FILE, 'rb') as fh:
            header = vdif.VDIFHeader.fromfile(fh)
            fh.seek(0)
            frameset = fh.read_frameset()

        assert len(frameset.frames) == 8
        assert len(frameset) == len(frameset.frames[0])
        assert frameset.samples_per_frame == 20000
        assert frameset.nchan == 1
        assert frameset.shape == (20000, 8, 1)
        assert frameset.size == 160000
        assert frameset.ndim == 3
        assert frameset.nbytes == 8 * frameset.frames[0].nbytes
        assert 'edv' in frameset
        assert 'edv' in frameset.keys()
        assert frameset['edv'] == 3
        # Properties from headers are passed on if they are settable.
        assert frameset.time == frameset.header0.time
        with pytest.raises(AttributeError):
            # But not all.
            frameset.update(1)
        assert ([fr.header['thread_id'] for fr in frameset.frames] ==
                list(range(8)))
        # Check frame associated with header makes sense.
        first_frame = frameset.frames[header['thread_id']]
        assert first_frame.header == header
        assert np.all(first_frame[:12, 0].astype(int) ==
                      np.array([1, 1, 1, -3, 1, 1, -3, -3, -3, 3, 3, -1]))
        # Check two other frames are in the right place.
        assert np.all(frameset.frames[0][:12, 0].astype(int) ==
                      np.array([-1, -1, 3, -1, 1, -1, 3, -1, 1, 3, -1, 1]))
        assert np.all(frameset.frames[3][:12, 0].astype(int) ==
                      np.array([-1, 1, -1, 1, -3, -1, 3, -1, 3, -3, 1, 3]))

        with vdif.open(SAMPLE_FILE, 'rb') as fh:
            frameset2 = fh.read_frameset(thread_ids=[2, 3])
            fh.fh_raw.seek(0)
            frameset3 = fh.read_frameset(thread_ids=[3, 4, 1])
        assert frameset2.shape == (20000, 2, 1)
        # Note: slicing of framesets themselves is checked below.
        assert np.all(frameset2.data == frameset.data[:, 2:4])
        assert np.all(frameset3.data == frameset.data[:, [3, 4, 1]])

        frameset3 = vdif.VDIFFrameSet(frameset.frames, frameset.header0)
        assert frameset3 == frameset
        frameset4 = vdif.VDIFFrameSet.fromdata(frameset.data, frameset.header0)
        assert np.all(frameset4.data == frameset.data)
        assert frameset4.time == frameset.time
        # the following check cannot be done with frameset itself, since
        # the times in four of the eight headers are wrong.
        frameset5 = vdif.VDIFFrameSet.fromdata(frameset.data,
                                               frameset4.header0)
        assert frameset5 == frameset4
        frameset6 = vdif.VDIFFrameSet.fromdata(frameset.data,
                                               **frameset4.header0)
        assert frameset6 == frameset4

        with vdif.open(SAMPLE_FILE, 'rb') as fh:
            fh.seek(0)
            # try reading just a few threads
            frameset4 = fh.read_frameset(thread_ids=[2, 3])
            assert frameset4.header0.time == frameset.header0.time
            assert np.all(frameset.data[:, 2:4] == frameset4.data)
            # Read beyond end
            fh.seek(-10064, 2)
            with pytest.raises(EOFError):
                fh.read_frameset(thread_ids=list(range(8)))
            # Give non-existent thread_id.
            fh.seek(0)
            with pytest.raises(IOError):
                fh.read_frameset(thread_ids=[1, 9])

    def test_frameset_getitem_setitem(self):
        with vdif.open(SAMPLE_FILE, 'rb') as fh:
            frameset = fh.read_frameset()

        # Try slicing a bit more, assuming whole-frameset read works.
        data = frameset.data
        # Whole frameset.
        assert np.all(frameset[()] == data)
        assert np.all(frameset[:] == data)
        # Select just samples.
        assert np.all(frameset[15] == data[15])
        assert np.all(frameset[10:20] == data[10:20])
        # Select just frames.
        assert np.all(frameset[:, 3] == data[:, 3])
        assert np.all(frameset[:, 2:4] == data[:, 2:4])
        # Select just channels (there is only one).
        assert np.all(frameset[:, :, 0] == data[:, :, 0])
        assert np.all(frameset[:, :, :1] == data[:, :, :1])
        # Check direct access to frames with known results.
        assert np.all(frameset[:12, 0, 0].astype(int) ==
                      np.array([-1, -1, 3, -1, 1, -1, 3, -1, 1, 3, -1, 1]))
        assert np.all(frameset[:12, 3, 0].astype(int) ==
                      np.array([-1, 1, -1, 1, -3, -1, 3, -1, 3, -3, 1, 3]))

        frameset2 = vdif.VDIFFrameSet.fromdata(data, frameset.header0)
        assert np.all(frameset2.data == data)
        frameset2[()] = 1.
        assert np.all(frameset2.data == 1.)
        frameset2[:] = data
        assert np.all(frameset2.data == data)
        # Select just samples.
        frameset2[15] = -data[15]
        assert np.all(frameset2[15] == -data[15])
        frameset2[10:20] = -1.
        assert np.all(frameset2[10:20] == -1.)
        frameset2[10:20:2] = data[10:20:2]
        assert np.all(frameset2[10:20:2] == data[10:20:2])
        assert np.all(frameset2[11:20:2] == -1.)
        # Select just frames.
        frameset2[:, 3] = -1
        assert np.all(frameset2[:, 3] == -1.)
        frameset2[:, 2:4] = 1.
        assert np.all(frameset2[:, 2:4] == 1.)
        frameset2[:, [0, 4, 5, 6]] = data[:, :4]
        frameset2[:, [1, 2, 3, 7]] = data[:, 4:]
        assert np.all(frameset2[:, [0, 4, 5, 6, 1, 2, 3, 7]] == data)
        # Select just channels (there is only one).
        frameset2[:, :, 0] = -data[:, :, 0]
        assert np.all(frameset2[:, :, 0] == -data[:, :, 0])
        frameset2[:, :, :1] = 1.
        assert np.all(frameset2[:, :, :1] == 1.)
        # Test header getting/setting.
        assert np.all(frameset2['thread_id'] == [f.header['thread_id']
                                                 for f in frameset2.frames])
        assert frameset2['frame_nr'] == frameset2.header0['frame_nr']
        frameset2['frame_nr'] = 25
        assert all(f.header['frame_nr'] == 25 for f in frameset2.frames)
        assert frameset2['frame_nr'] == 25
        frameset2['thread_id'] = np.arange(10, 18)
        assert all(f.header['thread_id'] == v
                   for f, v in zip(frameset2.frames, np.arange(10, 18)))
        assert all(frameset2['thread_id'] == np.arange(10, 18))
        with pytest.raises(ValueError):
            frameset2['thread_id'] = 0
        with pytest.raises(ValueError):
            frameset2['thread_id'] = 0, 1, 2, 3, 4, 5, 6, 1
        with pytest.raises(ValueError):
            frameset2['frame_nr'] = 0, 1, 0, 1, 0, 1, 0, 1
        # And a bit of getattr.
        assert frameset2.time == frameset2.header0.time
        assert frameset2.valid
        mixed_valid = True, True, False, False, True, True, False, False
        frameset2.valid = mixed_valid
        assert np.all(frameset2.valid == mixed_valid)
        frameset2.valid = True
        assert frameset2.valid
        frameset2.valid = False
        assert not frameset2.valid

    def test_find_header(self, tmpdir):
        # Below, the tests set the file pointer to very close to a header,
        # since otherwise they run *very* slow.  This is somehow related to
        # pytest, since speed is not a big issue running stuff on its own.
        with vdif.open(SAMPLE_FILE, 'rb') as fh:
            header0 = vdif.VDIFHeader.fromfile(fh)
            fh.seek(0)
            header_0 = fh.find_header(frame_nbytes=header0.frame_nbytes)
            assert fh.tell() == 0
            fh.seek(5000)
            header_5000f = fh.find_header(frame_nbytes=header0.frame_nbytes,
                                          forward=True)
            assert fh.tell() == header0.frame_nbytes
            # sample file has corrupted time in even threads; check this
            # doesn't matter
            fh.seek(15000)
            header_15000f = fh.find_header(frame_nbytes=header0.frame_nbytes,
                                           forward=True)
            assert fh.tell() == 3 * header0.frame_nbytes
            fh.seek(20128)
            header_20128f = fh.find_header(template_header=header0,
                                           forward=True)
            assert fh.tell() == 4 * header0.frame_nbytes
            fh.seek(16)
            header_16b = fh.find_header(frame_nbytes=header0.frame_nbytes,
                                        forward=False)
            assert fh.tell() == 0
            fh.seek(-10000, 2)
            header_m10000b = fh.find_header(frame_nbytes=header0.frame_nbytes,
                                            forward=False)
            assert fh.tell() == 14 * header0.frame_nbytes
            fh.seek(-5000, 2)
            header_m5000b = fh.find_header(frame_nbytes=header0.frame_nbytes,
                                           forward=False)
            assert fh.tell() == 15 * header0.frame_nbytes
            fh.seek(-20, 2)
            header_end = fh.find_header(template_header=header0,
                                        forward=True)
            assert header_end is None
        # thread order = 1,3,5,7,0,2,4,6
        assert header_16b == header_0
        # second frame
        assert header_5000f['frame_nr'] == 0
        assert header_5000f['thread_id'] == 3
        # fourth frame
        assert header_15000f['frame_nr'] == 0
        assert header_15000f['thread_id'] == 7
        # fifth frame
        assert header_20128f['frame_nr'] == 0
        assert header_20128f['thread_id'] == 0
        # one but last frame
        assert header_m10000b['frame_nr'] == 1
        assert header_m10000b['thread_id'] == 4
        # last frame
        assert header_m5000b['frame_nr'] == 1
        assert header_m5000b['thread_id'] == 6
        # Make file with missing data.
        with open(str(tmpdir.join('test.vdif')), 'w+b') as s, \
                open(SAMPLE_FILE, 'rb') as f:
            s.write(f.read(5100))
            f.seek(10000)
            s.write(f.read())
            with vdif.open(s, 'rb') as fh:
                fh.seek(0)
                header_0 = fh.find_header(template_header=header0)
                assert fh.tell() == 0
                fh.seek(5000)
                header_5000ft = fh.find_header(template_header=header0,
                                               forward=True)
                assert fh.tell() == header0.frame_nbytes * 2 - 4900
                header_5000f = fh.find_header(
                    frame_nbytes=header0.frame_nbytes, forward=True)
                assert fh.tell() == header0.frame_nbytes * 2 - 4900
        assert header_5000f['frame_nr'] == 0
        assert header_5000f['thread_id'] == 5
        assert header_5000ft == header_5000f
        # for completeness, also check a really short file...
        with open(str(tmpdir.join('test.vdif')), 'w+b') as s, \
                open(SAMPLE_FILE, 'rb') as f:
            s.write(f.read(5040))
            with vdif.open(s, 'rb') as fh:
                fh.seek(10)
                header_10 = fh.find_header(frame_nbytes=header0.frame_nbytes,
                                           forward=False)
                assert fh.tell() == 0
            assert header_10 == header0

    def test_filestreamer(self):
        with open(SAMPLE_FILE, 'rb') as fh:
            header = vdif.VDIFHeader.fromfile(fh)

        with vdif.open(SAMPLE_FILE, 'rs') as fh:
            assert fh.readable() is True
            assert fh.writable() is False
            assert fh.seekable() is True
            assert fh.closed is False
            assert repr(fh).startswith('<VDIFStreamReader')
            assert fh.tell() == 0
            assert header == fh.header0
            assert fh.sample_rate == 32*u.MHz
            assert fh.start_time == fh.header0.time
            assert abs(fh.time - fh.start_time) < 1. * u.ns
            assert fh.time == fh.tell(unit='time')
            record = fh.read(12)
            assert fh.tell() == 12
            t12 = fh.time
            s12 = 12 / fh.sample_rate
            assert abs(t12 - fh.start_time - s12) < 1. * u.ns
            fh.seek(10, 1)
            fh.tell() == 22
            fh.seek(t12)
            assert fh.tell() == 12
            fh.seek(-s12, 1)
            assert fh.tell() == 0
            with pytest.raises(ValueError):
                fh.seek(0, 3)
            # Test seeker works with both int and str values for whence
            assert fh.seek(13, 0) == fh.seek(13, 'start')
            assert fh.seek(-13, 2) == fh.seek(-13, 'end')
            fhseek_int = fh.seek(17, 1)
            fh.seek(-17, 'current')
            fhseek_str = fh.seek(17, 'current')
            assert fhseek_int == fhseek_str
            with pytest.raises(ValueError):
                fh.seek(0, 'last')
            assert fh.sample_shape == (8,)
            assert fh.shape == (40000,) + fh.sample_shape
            assert fh.size == np.prod(fh.shape)
            assert fh.ndim == len(fh.shape)
            assert abs(fh.stop_time - fh._last_header.time - (
                fh.samples_per_frame / fh.sample_rate)) < 1. * u.ns
            assert abs(fh.stop_time - fh.start_time -
                       (fh.shape[0] / fh.sample_rate)) < 1. * u.ns
            fh.seek(1, 'end')
            with pytest.raises(EOFError):
                fh.read()

        assert record.shape == (12, 8)
        assert np.all(record.astype(int)[:, 0] ==
                      np.array([-1, -1, 3, -1, 1, -1, 3, -1, 1, 3, -1, 1]))

        # Test that squeeze attribute works on read (including in-place read).
        with vdif.open(SAMPLE_FILE, 'rs') as fh:
            assert fh.sample_shape == (8,)
            assert fh.sample_shape.nthread == 8
            assert fh.read(1).shape == (1, 8)
            fh.seek(0)
            out_squeeze = np.zeros((12, 8))
            fh.read(out=out_squeeze)
            assert fh.tell() == 12
            assert np.all(out_squeeze == record)

        with vdif.open(SAMPLE_FILE, 'rs', squeeze=False) as fh:
            assert fh.sample_shape == (8, 1)
            assert fh.read(1).shape == (1, 8, 1)
            fh.seek(0)
            out_nosqueeze = np.zeros((12, 8, 1))
            fh.read(out=out_nosqueeze)
            assert fh.tell() == 12
            assert np.all(out_nosqueeze.squeeze() == out_squeeze)

    def test_stream_writer(self, tmpdir):
        vdif_file = str(tmpdir.join('simple.vdif'))
        # Try writing a very simple file, using edv=0.  With 16 samples per
        # frame and a sample_rate of 320 Hz, we have a frame rate of 20 Hz.
        # We start on purpose not on an integer second, to check that
        # frame numbers get set correctly automatically.
        start_time = Time('2010-11-12T13:14:15.25')
        # Same fake data for all frames.
        data = np.ones((16, 2, 2))
        data[5, 0, 0] = data[6, 1, 1] = -1.
        header = vdif.VDIFHeader.fromvalues(
            edv=0, time=start_time, nchan=2, bps=2,
            complex_data=False, thread_id=0, samples_per_frame=16,
            station='me', sample_rate=320*u.Hz)
        with vdif.open(vdif_file, 'ws', header0=header,
                       sample_rate=320*u.Hz, nthread=2) as fw:
            assert fw.sample_rate == 320*u.Hz
            for i in range(17):
                fw.write(data)
            # Write an invalid frame.
            fw.write(data, valid=False)
            # Write 3 frames using pieces.
            fw.write(data[:4])
            fw.write(np.concatenate((data[4:], data, data[:-4]), axis=0))
            fw.write(data[-4:])
            # Add 9 more just to fill it out to 30 frames.
            for i in range(9):
                fw.write(data)

        with vdif.open(vdif_file, 'rs') as fh:
            assert fh.header0.station == 'me'
            assert fh.samples_per_frame == 16
            assert fh.sample_rate == 320*u.Hz
            assert not fh.complex_data
            assert fh.header0.bps == 2
            assert fh.sample_shape.nchan == 2
            assert fh.sample_shape.nthread == 2
            assert fh.start_time == start_time
            assert np.abs(fh.stop_time - fh.start_time - 1.5 * u.s) < 1. * u.ns
            # Seek to 8 samples before the 17th frame (which has invalid data),
            # and check reading around that.
            fh.seek(16 * 17 - 8)
            record = fh.read(56)
            assert np.all(record[:8] == data[8:])
            assert np.all(record[8:24] == 0.)
            assert np.all(record[24:40] == data)
            assert np.all(record[40:] == data)

        # A bit random, but this is a good place to check that `info`
        # does the right thing for streams that do not start at frame nr 0.
        with vdif.open(vdif_file, 'rb') as fh:
            assert fh.info.frame_rate == 20. * u.Hz
            assert abs(fh.info.start_time - start_time) < 1. * u.ns

        # Test that failing to pass a sample rate returns a ValueError.
        # TODO: why can this not be determined from the frame rate?
        with pytest.raises(ValueError) as excinfo:
            with vdif.open(vdif_file, 'ws', header0=header,
                           nthread=2) as fw:
                pass
        assert "sample rate must be passed" in str(excinfo.value)

        # Test that squeeze attribute works on write.
        with vdif.open(SAMPLE_FILE, 'rs') as fh:
            record = fh.read()
            header = fh.header0

        test_file_squeeze = str(tmpdir.join('test_squeeze.vdif'))
        with vdif.open(test_file_squeeze, 'ws', header0=header,
                       nthread=8) as fws:
            assert fws.sample_shape == (8,)
            assert fws.sample_shape.nthread == 8
            fws.write(record)
        test_file_nosqueeze = str(tmpdir.join('test_nosqueeze.vdif'))
        with vdif.open(test_file_nosqueeze, 'ws', header0=header,
                       nthread=8, squeeze=False) as fwns:
            assert fwns.sample_shape == (8, 1)
            assert fwns.sample_shape.nthread == 8
            assert fwns.sample_shape.nchan == 1
            fwns.write(record[..., np.newaxis])

        with vdif.open(test_file_squeeze, 'rs') as fhs, \
                vdif.open(test_file_nosqueeze, 'rs') as fhns:
            assert np.all(fhs.read() == record)
            assert np.all(fhns.read() == record)

    def test_subset(self, tmpdir):
        # Use the default sample file.
        with vdif.open(SAMPLE_FILE, 'rs') as fh:
            data = fh.read()
            sample_rate = fh.sample_rate
            samples_per_frame = fh.samples_per_frame
            header0 = fh.header0

        with vdif.open(SAMPLE_FILE, 'rs', subset=slice(0, 8, 2)) as fhn:
            assert fhn.sample_shape == (4,)
            check = fhn.read()
            assert np.all(check == data[:, slice(0, 8, 2)])

        with vdif.open(SAMPLE_FILE, 'rs', subset=[0]) as fhn:
            assert fhn.sample_shape == (1,)
            check = fhn.read()
            assert check.shape == (data.shape[0], 1)
            assert np.all(check == data[:, :1])

        # Make an 8 channel file.
        test_file = str(tmpdir.join('test.vdif'))
        with vdif.open(test_file, 'ws', sample_rate=sample_rate,
                       samples_per_frame=samples_per_frame // 8, nthread=1,
                       nchan=8, complex_data=header0['complex_data'],
                       bps=header0.bps, edv=header0.edv,
                       station=header0.station, time=fh.start_time) as fw:
            fw.write(data)

        with vdif.open(test_file, 'rs') as fhn:
            assert np.all(fhn.read() == data)

        with vdif.open(SAMPLE_FILE, 'rs', subset=np.array([3, 7])) as fhn:
            assert fhn.sample_shape == (2,)
            check = fhn.read()
            assert np.all(check == data[:, np.array([3, 7])])

        with vdif.open(SAMPLE_FILE, 'rs', subset=[2]) as fhn:
            assert fhn.sample_shape == (1,)
            check = fhn.read()
            assert np.all(check == data[:, 2:3])

        # Make an 8 thread, 4 channel file.
        data4x = np.array([data, abs(data),
                           -data, -abs(data)]).transpose(1, 2, 0)
        with vdif.open(test_file, 'ws', sample_rate=sample_rate,
                       samples_per_frame=samples_per_frame // 4, nthread=8,
                       nchan=4, complex_data=header0['complex_data'],
                       bps=header0.bps, edv=header0.edv,
                       station=header0.station, time=fh.start_time) as fw:
            fw.write(data4x)

        # Sanity check by re-reading the file.
        with vdif.open(test_file, 'rs') as fhn:
            assert fhn.sample_shape == (8, 4)
            check = fhn.read()
            assert np.all(check == data4x)

        # Single thread and channel selection.
        with vdif.open(test_file, 'rs', subset=(6, 2)) as fhn:
            assert fhn.sample_shape == ()
            check = fhn.read()
            assert np.all(check == data4x[:, 6, 2])

        # Single thread, multi-channel selection.
        with vdif.open(test_file, 'rs', subset=(3, [1, 2])) as fhn:
            assert fhn.sample_shape == (2,)
            check = fhn.read()
            assert np.all(check == data4x[:, 3, 1:3])

        # Multi-thread, multi-channel selection
        subset_md = (np.array([5, 3])[:, np.newaxis], np.array([0, 2]))
        with vdif.open(test_file, 'rs', subset=subset_md) as fhn:
            assert fhn.sample_shape == (2, 2)
            thread_ids = [frame.header['thread_id'] for frame in
                          fhn._frameset.frames]
            assert thread_ids == [5, 3]
            check = fhn.read()
            assert np.all(check == data4x[(slice(None),) + subset_md])

    def test_stream_verify(self, tmpdir):
        """Test that we can pass or set verify=False to bypass header
        checking."""
        # Grab data from the sample file.
        with vdif.open(SAMPLE_FILE, 'rs') as fh:
            data = fh.read()

        testverifyfile = str(tmpdir.join('testverify.vdif'))
        # Make a file with a sync pattern error in the second set of frames.
        with vdif.open(SAMPLE_FILE, 'rb') as fh, \
                vdif.open(testverifyfile, 'wb') as fw:
            fr = fh.read_frameset()
            fr.tofile(fw)
            fr = fh.read_frameset()
            fr.frames[2].header.mutable = True
            fr.frames[2].header['sync_pattern'] = 0xabbaabba
            fr.tofile(fw)

        with vdif.open(testverifyfile, 'rs') as fn:
            assert fn.verify is True
            # This should fail at the second frameset.
            with pytest.raises(AssertionError):
                fn.read()
            assert fn.tell() == 20000
            fn.verify = False
            assert fn.verify is False
            assert np.all(fn.read() == data[20000:])

        # Check that we can pass verify=False.
        with vdif.open(testverifyfile, 'rs', verify=False) as fn:
            assert fn.verify is False
            assert np.all(fn.read() == data)

    # Test that writing an incomplete stream is possible, and that frame set is
    # appropriately marked as invalid.
    @pytest.mark.parametrize('fill_value', (0., -999.))
    def test_incomplete_stream(self, tmpdir, fill_value):
        vdif_incomplete = str(tmpdir.join('incomplete.vdif'))
        with catch_warnings(UserWarning) as w:
            with vdif.open(SAMPLE_FILE, 'rs') as fr:
                record = fr.read(10)
                with vdif.open(vdif_incomplete, 'ws', header0=fr.header0,
                               sample_rate=32*u.MHz, nthread=8) as fw:
                    fw.write(record)
        assert len(w) == 1
        assert 'partial buffer' in str(w[0].message)
        with vdif.open(vdif_incomplete, 'rs', fill_value=fill_value) as fwr:
            assert all([not frame.valid for frame in fwr._frameset.frames])
            assert fwr.fill_value == fill_value
            assert np.all(fwr.read() == fill_value)

    def test_corrupt_stream(self, tmpdir):
        with vdif.open(SAMPLE_FILE, 'rb') as fh, \
                open(str(tmpdir.join('test.vdif')), 'w+b') as s:
            frame = fh.read_frame()
            frame.tofile(s)
            # Now add lots of the next frame, i.e., with a different thread_id
            # and different frame_nr
            fh.seek(-frame.header.frame_nbytes, 2)
            frame2 = fh.read_frame()
            for i in range(15):
                frame2.tofile(s)
            s.seek(0)
            with vdif.open(s, 'rs') as f2:
                assert f2.header0 == frame.header
                with pytest.raises(ValueError):
                    f2._last_header

    def test_io_invalid(self, tmpdir):
        tmp_file = str(tmpdir.join('ts.dat'))
        with open(tmp_file, 'wb') as fw:
            fw.write(b'      ')

        with pytest.raises(TypeError):
            # Extra argument.
            vdif.open(tmp_file, 'rb', bla=10)
        with pytest.raises(ValueError):
            # Missing w or r.
            vdif.open(tmp_file, 's')


def test_vlbi_vdif():
    """Tests SAMPLE_VLBI, which is SAMPLE_VDIF with uncorrected timestamps."""
    with vdif.open(SAMPLE_VLBI, 'rs') as fh, \
            vdif.open(SAMPLE_FILE, 'rs') as fhc:
        assert fh.sample_rate == 32*u.MHz
        assert fh.start_time == fh.header0.time
        assert fh.start_time == fhc.start_time
        assert fh.shape == (40000,) + fh.sample_shape
        assert abs(fh.stop_time - fh._last_header.time - (
            fh.samples_per_frame / fh.sample_rate)) < 1. * u.ns
        assert abs(fh.stop_time - fh.start_time -
                   (fh.shape[0] / fh.sample_rate)) < 1. * u.ns
        assert np.all(fh.read() == fhc.read())


def test_mwa_vdif():
    """Test phased VDIF format (uses EDV=0)"""
    with vdif.open(SAMPLE_MWA, 'rs', sample_rate=1.28*u.MHz) as fh:
        assert fh.samples_per_frame == 128
        assert fh.sample_rate == 1.28*u.MHz
        assert fh.time == Time('2015-10-03T20:49:45.000')
        assert fh.header0.edv == 0


def test_arochime_vdif():
    """Test ARO CHIME format (uses EDV=0)"""
    # File has 1 sample/frame.
    frame_rate = sample_rate = 800*u.MHz / 1024. / 2.
    with open(SAMPLE_AROCHIME, 'rb') as fh:
        header0 = vdif.VDIFHeader.fromfile(fh)
    assert header0.edv == 0
    assert header0.samples_per_frame == 1
    assert header0['frame_nr'] == 308109
    with pytest.raises(ValueError):
        header0.time
    assert abs(header0.get_time(frame_rate=frame_rate) -
               Time('2016-04-22T08:45:31.788759040')) < 1. * u.ns
    # Also check writing Time.
    header1 = header0.copy()
    with pytest.raises(ValueError):
        header1.time = Time('2016-04-22T08:45:31.788759040')
    header1.set_time(Time('2016-04-22T08:45:32.788759040'),
                     frame_rate=frame_rate)
    assert abs(header1.get_time(frame_rate=frame_rate) -
               header0.get_time(frame_rate=frame_rate) - 1. * u.s) < 1.*u.ns

    # Now test the actual data stream.
    with vdif.open(SAMPLE_AROCHIME, 'rs', sample_rate=sample_rate) as fh:
        assert fh.samples_per_frame == 1
        t0 = fh.time
        assert abs(t0 - Time('2016-04-22T08:45:31.788759040')) < 1. * u.ns
        assert abs(t0 - fh.start_time) < 1. * u.ns
        assert fh.header0.edv == 0
        assert fh.shape == (5,) + fh.sample_shape
        d = fh.read()
        assert d.shape == (5, 2, 1024)
        assert d.dtype.kind == 'c'
        t1 = fh.time
        assert abs(t1 - fh.stop_time) < 1. * u.ns
        assert abs(t1 - t0 - fh.shape[0] / fh.sample_rate) < 1. * u.ns

    # For this file, we cannot find a frame rate, so opening it without
    # should fail.
    with pytest.raises(EOFError):
        with vdif.open(SAMPLE_AROCHIME, 'rs') as fh:
            pass


def test_legacy_vdif(tmpdir):
    """Create legacy header, ensuring it is not treated as EDV=0 (see #12)."""
    # legacy_mode, 1 sec, epoch 4, vdif v1, nchan=2, 507*8 bytes, bps=2, 'AA'
    words = (1 << 30 | 1, 4 << 24, 1 << 29 | 1 << 24 | 507, 1 << 26 | 0x4141)
    header = vdif.VDIFHeader(words)
    assert header['legacy_mode'] is True
    assert header.edv is False
    assert abs(header.time - Time('2002-01-01T00:00:01.000000000')) < 1. * u.ns
    assert header['frame_nr'] == 0
    assert header['vdif_version'] == 1
    assert header.nchan == 2
    assert header.frame_nbytes == 507 * 8
    assert header.nbytes == 16
    assert header['complex_data'] is False
    assert header.bps == 2
    assert header['thread_id'] == 0
    assert header.station == 'AA'
    with open(str(tmpdir.join('test.vdif')), 'w+b') as s:
        header.tofile(s)
        # Add fake payload
        s.write(np.zeros(503, dtype=np.int64).tostring())
        s.seek(0)
        header2 = vdif.VDIFHeader.fromfile(s)
    assert header2 == header



# Licensed under the GPLv3 - see LICENSE
from __future__ import division, unicode_literals, print_function

import numpy as np
from astropy.utils import lazyproperty
import astropy.units as u

from ..vlbi_base.base import (make_opener, VLBIFileBase, VLBIFileReaderBase,
                              VLBIStreamBase, VLBIStreamReaderBase,
                              VLBIStreamWriterBase)
from .header import Mark4Header
from .payload import Mark4Payload
from .frame import Mark4Frame
from .file_info import Mark4FileReaderInfo


__all__ = ['Mark4FileReader', 'Mark4FileWriter', 'Mark4StreamReader',
           'Mark4StreamWriter', 'open']

# Look-up table for the number of bits in a byte.
nbits = ((np.arange(256)[:, np.newaxis] >> np.arange(8) & 1)
         .sum(1).astype(np.int16))


class Mark4FileReader(VLBIFileReaderBase):
    """Simple reader for Mark 4 files.

    Wraps a binary filehandle, providing methods to help interpret the data,
    such as `locate_frame, `read_frame` and `get_frame_rate`.

    Parameters
    ----------
    fh_raw : filehandle
        Filehandle of the raw binary data file.
    ntrack : int or None, optional.
        Number of Mark 4 bitstreams.  Can be determined automatically as
        part of locating the first frame.
    decade : int or None
        Decade in which the observations were taken.  Can instead pass an
        approximate ``ref_time``.
    ref_time : `~astropy.time.Time` or None
        Reference time within 4 years of the observation time.  Used only
        if ``decade`` is not given.
    """
    def __init__(self, fh_raw, ntrack=None, decade=None, ref_time=None):
        self.ntrack = ntrack
        self.decade = decade
        self.ref_time = ref_time
        super(Mark4FileReader, self).__init__(fh_raw)

    def __repr__(self):
        return ("{name}(fh_raw={s.fh_raw}, ntrack={s.ntrack}, "
                "decade={s.decade}, ref_time={s.ref_time})"
                .format(name=self.__class__.__name__, s=self))

    info = Mark4FileReaderInfo()

    def read_header(self):
        """Read a single header from the file.

        Returns
        -------
        header : `~baseband.mark4.Mark4Header`
        """
        return Mark4Header.fromfile(self, ntrack=self.ntrack,
                                    decade=self.decade, ref_time=self.ref_time)

    def read_frame(self, verify=True):
        """Read a single frame (header plus payload).

        Returns
        -------
        frame : `~baseband.mark4.Mark4Frame`
            With ``.header`` and ``.data`` properties that return the
            :class:`~baseband.mark4.Mark4Header` and data encoded in the frame,
            respectively.
        verify : bool, optional
            Whether to do basic checks of frame integrity.  Default: `True`.
        """
        return Mark4Frame.fromfile(self.fh_raw, self.ntrack,
                                   decade=self.decade, ref_time=self.ref_time,
                                   verify=verify)

    def get_frame_rate(self):
        """Determine the number of frames per second.

        The frame rate is calculated from the time elapsed between the
        first two frames, as inferred from their time stamps.

        Returns
        -------
        frame_rate : `~astropy.units.Quantity`
            Frames per second.
        """
        oldpos = self.tell()
        self.seek(0)
        try:
            self.locate_frame()
            header0 = self.read_header()
            self.seek(header0.payload_nbytes, 1)
            header1 = self.read_header()
        finally:
            self.seek(oldpos)

        # Mark 4 specification states frames-lengths range from 1.25 ms
        # to 160 ms.
        tdelta = header1.fraction[0] - header0.fraction[0]
        return np.round(1 / tdelta) * u.Hz

    def locate_frame(self, forward=True, maximum=None):
        """Locate the frame nearest the current position.

        The search is for the following pattern:

        * 32*tracks bits set at offset bytes
        * 1*tracks bits unset before offset
        * 32*tracks bits set at offset+2500*tracks bytes

        This reflects 'sync_pattern' of 0xffffffff for a given header and one
        a frame ahead, which is in word 2, plus the lsb of word 1, which is
        'system_id'.

        If the file does not have ntrack is set, it will be auto-determined.

        Parameters
        ----------
        forward : bool, optional
            Whether to search forwards or backwards.  Default: `True`.
        maximum : int, optional
            Maximum number of bytes forward to search through.
            Default: twice the frame size (``20000 * ntrack // 8``).

        Returns
        -------
        offset : int or `None`
            Byte offset of the next frame. `None` if the search was not
            successful.
        """
        fh = self.fh_raw
        file_pos = fh.tell()
        # Use initializer value (determines ntrack if not already given).
        ntrack = self.ntrack
        if ntrack is None:
            fh.seek(0)
            ntrack = self.determine_ntrack(maximum=maximum)
            if ntrack is None:
                raise ValueError("cannot determine ntrack automatically. "
                                 "Try passing in an explicit value.")
            if forward and fh.tell() >= file_pos:
                return fh.tell()

            fh.seek(file_pos)

        nset = np.ones(32 * ntrack // 8, dtype=np.int16)
        nunset = np.ones(ntrack // 8, dtype=np.int16)
        frame_nbytes = ntrack * 2500
        fh.seek(0, 2)
        filesize = fh.tell()
        if filesize < frame_nbytes:
            fh.seek(file_pos)
            return None

        if maximum is None:
            maximum = 2 * frame_nbytes
        # Loop over chunks to try to find the frame marker.
        step = frame_nbytes // 2
        # Read a bit more at every step to ensure we don't miss a "split"
        # header.
        block = step + 160 * ntrack // 8
        if forward:
            iterate = range(max(min(file_pos, filesize - block), 0),
                            max(min(file_pos + maximum, filesize - block + 1),
                                1),
                            step)
        else:
            iterate = range(min(max(file_pos - step, 0), filesize - block),
                            min(max(file_pos - step - maximum - 1, -1),
                                filesize - block),
                            -step)
        for frame in iterate:
            fh.seek(frame)

            data = np.frombuffer(fh.read(block), dtype=np.uint8)
            assert len(data) == block
            # Find header pattern.
            databits1 = nbits[data]
            nosync = np.convolve(databits1[len(nunset):] < 6, nset, 'valid')
            nolow = np.convolve(databits1[:-len(nset)] > 1, nunset, 'valid')
            wrong = nosync + nolow
            possibilities = np.where(wrong == 0)[0]
            # Check candidates by seeing whether there is a sync word
            # a frame size ahead. (Note: loop can be empty.)
            for possibility in possibilities[::1 if forward else -1]:
                # Real start of possible header.
                frame_start = frame + possibility - 63 * ntrack // 8
                if (forward and frame_start < file_pos or
                        not forward and frame_start > file_pos):
                    continue
                # Check there is a header following this.
                check = frame_start + frame_nbytes
                if check >= filesize - 32 * 2 * ntrack // 8 - len(nunset):
                    # But do before this one if we're beyond end of file.
                    check = frame_start - frame_nbytes
                    if check < 0:  # Assume OK if only one frame fits in file.
                        if frame_start + frame_nbytes > filesize:
                            continue
                        else:
                            break

                fh.seek(check + 32 * 2 * ntrack // 8)
                check_data = np.frombuffer(fh.read(len(nunset)),
                                           dtype=np.uint8)
                databits2 = nbits[check_data]
                if np.all(databits2 >= 6):
                    break  # Got it!

            else:  # None of them worked, so do next block.
                continue

            fh.seek(frame_start)
            return frame_start

        fh.seek(file_pos)
        return None

    def determine_ntrack(self, maximum=None):
        """Determines the number of tracks, by seeking the next frame.

        Uses `find_frame` to look for the first occurrence of a frame from
        the current position for all supported ``ntrack`` values.  Returns the
        first ``ntrack`` for which `find_frame` is successful, setting
        the file's ``ntrack`` property appropriately, and leaving the
        file pointer at the start of the frame.

        Parameters
        ----------
        maximum : int, optional
            Maximum number of bytes forward to search through.
            Default: twice the frame size (``20000 * ntrack // 8``).

        Returns
        -------
        ntrack : int or None
            Number of Mark 4 bitstreams.  `None` if no frame was found.
        """
        # Currently only 16, 32 and 64-track frames supported.
        old_ntrack = self.ntrack
        for ntrack in 16, 32, 64:
            try:
                self.ntrack = ntrack
                if self.locate_frame(maximum=maximum) is not None:
                    return ntrack
            except Exception:
                self.ntrack = old_ntrack
                raise

        self.ntrack = old_ntrack
        return None

    def find_header(self, forward=True, maximum=None):
        """Find the nearest header from the current position.

        If successful, the file pointer is left at the start of the header.

        Parameters
        ----------
        forward : bool, optional
            Seek forward if `True` (default), backward if `False`.
        maximum : int, optional
            Maximum number of bytes forward to search through.
            Default: twice the frame size (``20000 * ntrack // 8``).

        Returns
        -------
        header : :class:`~baseband.mark4.Mark4Header` or None
            Retrieved Mark 4 header, or `None` if nothing found.
        """
        offset = self.locate_frame(forward=forward)
        if offset is None:
            return None
        header = self.read_header()
        self.fh_raw.seek(offset)
        return header


class Mark4FileWriter(VLBIFileBase):
    """Simple writer for Mark 4 files.

    Adds `write_frame` method to the VLBI binary file wrapper.
    """

    def write_frame(self, data, header=None, **kwargs):
        """Write a single frame (header plus payload).

        Parameters
        ----------
        data : `~numpy.ndarray` or `~baseband.mark4.Mark4Frame`
            If an array, a header should be given, which will be used to
            get the information needed to encode the array, and to construct
            the Mark 4 frame.
        header : `~baseband.mark4.Mark4Header`
            Can instead give keyword arguments to construct a header.  Ignored
            if payload is a :class:`~baseband.mark4.Mark4Frame` instance.
        **kwargs :
            If ``header`` is not given, these are used to initialize one.
        """
        if not isinstance(data, Mark4Frame):
            data = Mark4Frame.fromdata(data, header, **kwargs)
        return data.tofile(self.fh_raw)


class Mark4StreamBase(VLBIStreamBase):
    """Base for Mark 4 streams."""

    def __init__(self, fh_raw, header0, sample_rate=None, squeeze=True,
                 subset=(), fill_value=0., verify=True):
        super(Mark4StreamBase, self).__init__(
            fh_raw, header0=header0, sample_rate=sample_rate,
            samples_per_frame=header0.samples_per_frame,
            unsliced_shape=(header0.nchan,),
            bps=header0.bps, complex_data=False, squeeze=squeeze,
            subset=subset, fill_value=fill_value, verify=verify)
        self._frame_rate = int(round((self.sample_rate /
                                      self.samples_per_frame).to_value(u.Hz)))


class Mark4StreamReader(Mark4StreamBase, VLBIStreamReaderBase):
    """VLBI Mark 4 format reader.

    Allows access to a Mark 4 file as a continuous series of samples.  Parts
    of the data stream replaced by header values are filled in.

    Parameters
    ----------
    fh_raw : filehandle
        Filehandle of the raw Mark 4 stream.
    sample_rate : `~astropy.units.Quantity`, optional
        Number of complete samples per second, i.e. the rate at which each
        channel is sampled.  If `None`, will be inferred from scanning two
        frames of the file.
    ntrack : int or None, optional
        Number of Mark 4 bitstreams.  If `None` (default), will attempt to
        automatically detect it by scanning the file.
    decade : int or None
        Decade of the observation start time (eg. ``2010`` for 2018), needed to
        remove ambiguity in the Mark 4 time stamp.  Can instead pass an
        approximate ``ref_time``.
    ref_time : `~astropy.time.Time` or None
        Reference time within 4 years of the start time of the observations.
        Used only if ``decade`` is not given.
    squeeze : bool, optional
        If `True` (default), remove any dimensions of length unity from
        decoded data.
    subset : indexing object, optional
        Specific channels of the complete sample to decode (after possible
        squeezing).  If an empty tuple (default), all channels are read.
    fill_value : float or complex, optional
        Value to use for invalid or missing data. Default: 0.
    verify : bool, optional
        Whether to do basic checks of frame integrity when reading.  The first
        frame of the stream is always checked.  Default: `True`.
    """

    _sample_shape_maker = Mark4Payload._sample_shape_maker

    def __init__(self, fh_raw, sample_rate=None, ntrack=None, decade=None,
                 ref_time=None, squeeze=True, subset=(), fill_value=0.,
                 verify=True):

        if decade is None and ref_time is None:
            raise TypeError("Mark 4 stream reader requires either decade or "
                            "ref_time to be passed in.")

        # Get binary file reader.
        fh_raw = Mark4FileReader(fh_raw, ntrack=ntrack, decade=decade,
                                 ref_time=ref_time)
        # Find first header, determining ntrack if needed.
        header0 = fh_raw.find_header()
        assert header0 is not None, (
            "could not find a first frame using ntrack={}. Perhaps "
            "try ntrack=None for auto-determination.".format(ntrack))
        self._offset0 = fh_raw.tell()
        super(Mark4StreamReader, self).__init__(
            fh_raw, header0=header0, sample_rate=sample_rate,
            squeeze=squeeze, subset=subset, fill_value=fill_value,
            verify=verify)
        # Use reference time in preference to decade so that a stream wrapping
        # a decade will work.
        self.fh_raw.decade = None
        self.fh_raw.ref_time = self.start_time

    @lazyproperty
    def _last_header(self):
        """Last header of the file."""
        last_header = super(Mark4StreamReader, self)._last_header
        # Infer the decade, assuming the end of the file is no more than
        # 4 years away from the start.
        last_header.infer_decade(self.start_time)
        return last_header

    def _read_frame(self, index):
        self.fh_raw.seek(self._offset0 + index * self.header0.frame_nbytes)
        frame = self.fh_raw.read_frame(verify=self.verify)
        # Set decoded value for invalid data.
        frame.fill_value = self.fill_value
        # TODO: add check that we got the right frame.
        return frame


class Mark4StreamWriter(Mark4StreamBase, VLBIStreamWriterBase):
    """VLBI Mark 4 format writer.

    Encodes and writes sequences of samples to file.

    Parameters
    ----------
    raw : filehandle
        Which will write filled sets of frames to storage.
    header0 : `~baseband.mark4.Mark4Header`
        Header for the first frame, holding time information, etc.  Can instead
        give keyword arguments to construct a header (see ``**kwargs``).
    sample_rate : `~astropy.units.Quantity`
        Number of complete samples per second, i.e. the rate at which each
        channel is sampled.  Needed to calculate header timestamps.
    squeeze : bool, optional
        If `True` (default), `write` accepts squeezed arrays as input, and
        adds any dimensions of length unity.
    **kwargs
        If no header is given, an attempt is made to construct one from these.
        For a standard header, this would include the following.

    --- Header keywords : (see :meth:`~baseband.mark4.Mark4Header.fromvalues`)

    time : `~astropy.time.Time`
        Start time of the file.  Sets bcd-encoded unit year, day, hour, minute,
        second in the header.
    ntrack : int
        Number of Mark 4 bitstreams (equal to number of channels times
        ``fanout`` times ``bps``)
    bps : int
        Bits per elementary sample.
    fanout : int
        Number of tracks over which a given channel is spread out.
    """

    _sample_shape_maker = Mark4Payload._sample_shape_maker

    def __init__(self, fh_raw, header0=None, sample_rate=None, squeeze=True,
                 **kwargs):
        if header0 is None:
            header0 = Mark4Header.fromvalues(**kwargs)
        super(Mark4StreamWriter, self).__init__(
            fh_raw=fh_raw, header0=header0, sample_rate=sample_rate,
            squeeze=squeeze)
        # Set up initial payload with right shape.
        samples_per_payload = (
            header0.samples_per_frame * header0.payload_nbytes //
            header0.frame_nbytes)
        self._payload = Mark4Payload.fromdata(
            np.zeros((samples_per_payload, header0.nchan), np.float32),
            header0)

    def _make_frame(self, frame_index):
        header = self.header0.copy()
        header.update(time=self.start_time + frame_index /
                      self._frame_rate * u.s)
        # Reuse payload.
        return Mark4Frame(header, self._payload)


open = make_opener('Mark4', globals(), doc="""
--- For reading a stream : (see `~baseband.mark4.base.Mark4StreamReader`)

sample_rate : `~astropy.units.Quantity`, optional
    Number of complete samples per second, i.e. the rate at which each channel
    is sampled.  If not given, will be inferred from scanning two frames of
    the file.
ntrack : int, optional
    Number of Mark 4 bitstreams.  If `None` (default), will attempt to
    automatically detect it by scanning the file.
decade : int or None
    Decade of the observation start time (eg. ``2010`` for 2018), needed to
    remove ambiguity in the Mark 4 time stamp (default: `None`).  Can instead
    pass an approximate ``ref_time``.
ref_time : `~astropy.time.Time` or None
    Reference time within 4 years of the start time of the observations.  Used
    only if ``decade`` is not given.
squeeze : bool, optional
    If `True` (default), remove any dimensions of length unity from
    decoded data.
subset : indexing object, optional
    Specific channels of the complete sample to decode (after possible
    squeezing).  If an empty tuple (default), all channels are read.
fill_value : float or complex, optional
    Value to use for invalid or missing data. Default: 0.
verify : bool, optional
    Whether to do basic checks of frame integrity when reading.  The first
    frame of the stream is always checked.  Default: `True`.

--- For writing a stream : (see `~baseband.mark4.base.Mark4StreamWriter`)

header0 : `~baseband.mark4.Mark4Header`
    Header for the first frame, holding time information, etc.  Can instead
    give keyword arguments to construct a header (see ``**kwargs``).
sample_rate : `~astropy.units.Quantity`
    Number of complete samples per second, i.e. the rate at which each channel
    is sampled.  Needed to calculate header timestamps.
squeeze : bool, optional
    If `True` (default), writer accepts squeezed arrays as input, and adds
    any dimensions of length unity.
**kwargs
    If the header is not given, an attempt will be made to construct one
    with any further keyword arguments.  See
    :class:`~baseband.mark4.base.Mark4StreamWriter`.

Returns
-------
Filehandle
    :class:`~baseband.mark4.base.Mark4FileReader` or
    :class:`~baseband.mark4.base.Mark4FileWriter` (binary), or
    :class:`~baseband.mark4.base.Mark4StreamReader` or
    :class:`~baseband.mark4.base.Mark4StreamWriter` (stream)
""")



# Licensed under the GPLv3 - see LICENSE
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import io
import itertools
from bisect import bisect
import numpy as np
from astropy.utils import lazyproperty

__all__ = ['SequentialFileReader', 'SequentialFileWriter', 'open']


class SequentialFileBase(object):
    """Deal with several files as if they were one contiguous one.

    For details, see `SequentialFileReader` and `SequentialFileWriter`.
    """
    def __init__(self, files, mode='rb', opener=None):
        self.files = files
        self.mode = mode
        self.opener = io.open if opener is None else opener
        self.file_nr = None
        self._file_sizes = []
        self._file_offsets = [0]
        self._open(0)

    def __getattr__(self, attr):
        """Try to get things on the current open file if it is not on self."""
        if not attr.startswith('_'):
            try:
                return getattr(self.fh, attr)
            except AttributeError:
                pass
        return self.__getattribute__(attr)

    def _open(self, file_nr):
        """Open the ``file_nr``th file of the list of underlying files.

        If a different file was already open, it is closed.  Nothing is done
        if the requested file is already open.
        """
        if file_nr != self.file_nr:
            try:
                fh = self.opener(self.files[file_nr], mode=self.mode)
            except IndexError:
                raise OSError('ran out of files.')
            if self.file_nr is not None:
                self.fh.close()
            self.fh = fh
            self.file_nr = file_nr
            if self.file_nr == len(self._file_sizes):
                file_size = self.file_size
                if file_size is not None:  # can happen for single-file write.
                    self._file_sizes.append(file_size)
                    self._file_offsets.append(self._file_offsets[-1] +
                                              file_size)

    def tell(self):
        """Return the current stream position."""
        return self._file_offsets[self.file_nr] + self.fh.tell()

    def memmap(self, dtype=np.uint8, mode=None, offset=None, shape=None,
               order='C'):
        """Map part of the file in memory.

        Note that the map cannnot span multiple underlying files.
        Parameters are as for `~numpy.memmap`.
        """
        if self.closed:
            raise ValueError('memmap of closed file.')

        dtype = np.dtype(dtype)

        if mode is None:
            mode = self.mode.replace('b', '')

        if offset is not None and offset != self.tell():
            # seek will fail for SequentialFileWriter, so we try to avoid it.
            self.seek(offset)
        elif self.fh.tell() == self._file_sizes[self.file_nr]:
            self._open(self.file_nr + 1)

        if shape is None:
            count = self.size - self.tell()
            if count % dtype.itemsize:
                raise ValueError("size of available data is not a "
                                 "multiple of the data-type size.")
            shape = (count // dtype.itemsize,)
        else:
            if not isinstance(shape, tuple):
                shape = (shape,)
            count = dtype.itemsize
            for k in shape:
                count *= k

        if self.fh.tell() + count > self._file_sizes[self.file_nr]:
            raise ValueError('mmap length exceeds individual file size')

        file_offset = self.fh.tell()
        mm = np.memmap(self.fh, dtype, mode, file_offset, shape, order)
        self.fh.seek(file_offset + count)
        return mm

    def close(self):
        """Close the currently open local file, and therewith the set."""
        if self.file_nr is not None:
            self.fh.close()
            self.file_nr = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __repr__(self):
        base_repr = ("{0}(files={1}, mode='{2}')"
                     .format(self.__class__.__name__, self.files, self.mode))
        extra = ("At offset: {0}; open file: {1!r}."
                 .format(self.tell(), None if self.file_nr is None else
                         self.files[self.file_nr]))
        return base_repr + "\n# " + extra


class SequentialFileReader(SequentialFileBase):
    """Read several files as if they were one contiguous one.

    Parameters
    ----------
    files : list, tuple, or other iterable of str, filehandle
        The contains the names of the underlying files that should be combined.
        If not a list or tuple, it should allow indexing with positive indices,
        and raise `IndexError` if these are out of range.
    mode : str, optional
        The mode with which the files should be opened (default: 'rb')
    opener : callable, optional
        Function to open a single file (default: `io.open`).
    """

    @property
    def file_size(self):
        """Size of the underlying file currently open for reading."""
        offset = self.fh.tell()
        file_size = self.fh.seek(0, 2)
        self.fh.seek(offset)
        return file_size

    @lazyproperty
    def size(self):
        """Size of all underlying files combined."""
        offset = self.tell()
        for i in itertools.count(start=len(self._file_sizes)):
            try:
                self._open(i)
            except Exception:
                break

        self.seek(offset)
        return self._file_offsets[-1]

    def seek(self, offset, whence=0):
        if self.closed:
            raise ValueError('seek of closed file.')

        if whence == 1:
            offset += self.tell()
        elif whence == 2:
            offset += self.size
        elif whence != 0:
            raise ValueError("invalid 'whence'; should be 0, 1, or 2.")

        if offset < 0:
            raise OSError('invalid offset')

        # If the offset is not in the current file, find right one.
        while not (0 <= offset - self._file_offsets[self.file_nr] <
                   self._file_sizes[self.file_nr]):
            # Note that not all files may have been opened at this point.
            # In that case, bisecting would find we're out of the current files
            # and one would open a new one.  The while loop ensures we keep
            # trying until we've got there or reached the end of the files.
            file_nr = bisect(self._file_offsets, offset) - 1
            try:
                self._open(file_nr)
            except (OSError, IOError):
                # If no files left, put pointer beyond end of last file.
                if file_nr != len(self._file_sizes):  # pragma: no cover
                    raise
                self._open(file_nr - 1)
                break

        self.fh.seek(offset - self._file_offsets[self.file_nr])
        return self.tell()
    seek.__doc__ = io.BufferedIOBase.seek.__doc__

    def read(self, count=None):
        if self.closed:
            raise ValueError('read of closed file.')

        if count is None or count < 0:
            count = max(self.size - self.tell(), 0)

        data = b''
        while count > 0:
            extra = self.fh.read(count)
            if not extra:
                break
            count -= len(extra)
            if not data:  # avoid copies for first read.
                data = extra
            else:
                data += extra
            # Go to current offset, possibly opening new file.
            self.seek(0, 1)

        return data
    read.__doc__ = io.BufferedIOBase.read.__doc__


class SequentialFileWriter(SequentialFileBase):
    """Write several files as if they were one contiguous one.

    Note that the file is not seekable and readable.

    Parameters
    ----------
    files : list, tuple, or other iterable of str, filehandle
        The contains the names of the underlying files that should be combined.
        If not a list or tuple, it should allow indexing with positive indices
        (e.g., returning a name as derived from a template).  It should raise
        raise `IndexError` if the index is out of range.
    mode : str, optional
        The mode with which the files should be opened (default: 'w+b'). If
        this does not include '+' for reading, memory maps are not possibe.
    file_size : int, optional
        The maximum size a file is allowed to have.  Default: `None`, which
        means it is unlimited and only a single file will be written (making
        using this class somewhat pointless).
    opener : callable, optional
        Function to open a single file (default: `io.open`).
    """
    def __init__(self, files, mode='w+b', file_size=None, opener=None):
        self.file_size = file_size
        super(SequentialFileWriter, self).__init__(files, mode, opener)

    def write(self, data):
        if self.closed:
            raise ValueError('write to closed file.')
        offset0 = self.tell()
        if self.file_size is not None:
            remaining = self.file_size - self.fh.tell()
            while len(data) > remaining:
                self.fh.write(data[:remaining])
                data = data[remaining:]
                self._open(self.file_nr + 1)
                remaining = self.file_size

        self.fh.write(data)
        return self.tell() - offset0
    write.__doc__ = io.BufferedIOBase.write.__doc__

    def memmap(self, dtype=np.uint8, mode=None, offset=None, shape=None,
               order='C'):
        """Map part of the file in memory.  Cannnot span file boundaries."""
        if shape is None:
            raise ValueError('cannot make writable memmap without shape.')
        return super(SequentialFileWriter,
                     self).memmap(dtype, mode, offset, shape, order)


def open(files, mode='rb', file_size=None, opener=None):
    """Read or write several files as if they were one contiguous one.

    Parameters
    ----------
    files : list, tuple, or other iterable of str, filehandle
        Contains the names of the underlying files that should be combined,
        ordered in time.  If not a list or tuple, it should allow indexing with
        positive indices, and raise `IndexError` if these are out of range.
    mode : str, optional
        The mode with which the files should be opened (default: 'rb').
    file_size : int, optional
        For writing, the maximum size of a file, beyond which a new file should
        be opened.  Default: `None`, which means it is unlimited and only a
        single file will be written.
    opener : callable, optional
        Function to open a single file (default: `io.open`).

    Notes
    -----
    The returned reader/writer will have a ``memmap`` method with which part of
    the files can be mapped to memory (like with `~numpy.memmap`), as long as
    those parts do not span files (and the underlying files are regular ones).
    For writing, this requires opening in read-write mode (i.e., 'w+b').

    Methods other than ``read``, ``write``, ``seek``, ``tell``, and ``close``
    are tried on the underlying file.  This implies, e.g., ``readline`` is
    possible, though the line cannot span multiple files.

    The reader assumes the sequence of files is **contiguous in time**, ie.
    with no gaps in the data.
    """
    if 'r' in mode:
        if file_size is not None:
            raise TypeError("cannot pass in 'file_size' for reading.")
        return SequentialFileReader(files, mode, opener=opener)
    elif 'w' in mode:
        return SequentialFileWriter(files, mode, file_size=file_size,
                                    opener=opener)
    else:
        raise ValueError("invalid mode '{0}'".format(mode))



# Licensed under the GPLv3 - see LICENSE
"""
Definitions for VLBI VDIF Headers.

Implements a VDIFHeader class used to store header words, and decode/encode
the information therein.

For the VDIF specification, see http://www.vlbi.org/vdif
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import numpy as np
import astropy.units as u
from astropy.extern import six

from astropy.time import Time, TimeDelta

from ..vlbi_base.header import (four_word_struct, eight_word_struct,
                                HeaderParser, VLBIHeaderBase)
from ..mark5b.header import Mark5BHeader


__all__ = ['VDIFHeader', 'VDIFBaseHeader', 'VDIFSampleRateHeader',
           'VDIFLegacyHeader', 'VDIFHeader0', 'VDIFHeader1',
           'VDIFHeader2', 'VDIFHeader3', 'VDIFMark5BHeader',
           'VDIF_HEADER_CLASSES']


ref_max = int(2. * (Time.now().jyear - 2000.)) + 1
ref_epochs = Time(['{y:04d}-{m:02d}-01'.format(y=2000 + ref // 2,
                                               m=1 if ref % 2 == 0 else 7)
                   for ref in range(ref_max)], format='isot', scale='utc',
                  precision=9)


VDIF_HEADER_CLASSES = {}
"""Dict for storing VDIF header class definitions, indexed by their EDV."""


class VDIFHeaderMeta(type):
    """
    Registry of VDIF Header EDV types, using the ``VDIF_HEADER_CLASSES``
    dict.  Checks for keyword and subclass conflicts before registering.

    This metaclass automatically registers any subclass of
    `~baseband.vdif.VDIFHeader`. This feature can be used to insert new EDV
    types into ``VDIF_HEADER_CLASSES``; see the :ref:`New VDIF EDV <new_edv>`
    tutorial.
    """
    _registry = VDIF_HEADER_CLASSES

    def __init__(cls, name, bases, dct):

        # Ignore VDIFHeader and VDIFBaseHeader, register others
        if name not in ('VDIFHeader', 'VDIFBaseHeader',
                        'VDIFSampleRateHeader'):

            # Extract edv from class; convert to -1 if edv is False
            # for VDIFLegacy
            edv = cls._edv
            if edv is False:
                edv = -1

            # Check if EDV is already registered (edv == None is invalid).
            if edv is None:
                raise ValueError("EDV cannot be None.  It should be "
                                 "overridden by the subclass.")
            elif edv in VDIFHeaderMeta._registry:
                raise ValueError("EDV {0} already registered in "
                                 "VDIF_HEADER_CLASSES".format(edv))

            VDIFHeaderMeta._registry.update({edv: cls})

        # If header parser has a sync pattern, append it as a private
        # attribute for cls.verify.
        if (hasattr(cls, '_header_parser') and
                'sync_pattern' in cls._header_parser.keys()):
            cls._sync_pattern = cls._header_parser.defaults['sync_pattern']

        super(VDIFHeaderMeta, cls).__init__(name, bases, dct)


@six.add_metaclass(VDIFHeaderMeta)
class VDIFHeader(VLBIHeaderBase):
    """VDIF Header, supporting different Extended Data Versions.

    Will initialize a header instance appropriate for a given EDV.
    See http://www.vlbi.org/vdif/docs/VDIF_specification_Release_1.1.1.pdf

    Parameters
    ----------
    words : tuple of int, or None
        Eight (or four for legacy VDIF) 32-bit unsigned int header words.
        If `None`, set to a tuple of zeros for later initialisation.
    edv : int, False, or None, optional
        Extended data version.  If `False`, a legacy header is used.
        If `None` (default), it is determined from the header.  (Given it
        explicitly is mostly useful for a slight speed-up.)
    verify : bool
        Whether to do basic verification of integrity.  Default: `True`.

    Returns
    -------
    header : `~baseband.vdif.VDIFHeader` subclass
        As appropriate for the extended data version.
    """

    _properties = ('frame_nbytes', 'payload_nbytes', 'bps', 'nchan',
                   'samples_per_frame', 'station', 'time')
    """Properties accessible/usable in initialisation for all VDIF headers."""

    _edv = None
    _struct = eight_word_struct

    def __new__(cls, words, edv=None, verify=True, **kwargs):
        # We use edv to define which class we return.
        if edv is None:
            # If not given, we extract edv from the header words.  This uses
            # parsers defined below, in VDIFBaseHeader.
            base_parsers = VDIFBaseHeader._header_parser.parsers
            if base_parsers['legacy_mode'](words):
                edv = False
            else:
                edv = base_parsers['edv'](words)

        # Have to use key "-1" instead of "False" since the dict-lookup treats
        # 0 and False as identical.
        cls = VDIF_HEADER_CLASSES.get(edv if edv is not False else -1,
                                      VDIFBaseHeader)
        return super(VDIFHeader, cls).__new__(cls)

    def __init__(self, words, edv=None, verify=True, **kwargs):
        if edv is not None:
            self._edv = edv
        super(VDIFHeader, self).__init__(words, verify=verify, **kwargs)

    def copy(self):
        return super(VDIFHeader, self).copy(edv=self.edv)

    def same_stream(self, other):
        """Whether header is consistent with being from the same stream."""
        # EDV and most parts of words 2 and 3 should be invariant.
        return (self.edv == other.edv and
                all(self[key] == other[key]
                    for key in ('ref_epoch', 'vdif_version', 'frame_length',
                                'complex_data', 'bits_per_sample',
                                'station_id')))

    @classmethod
    def fromfile(cls, fh, edv=None, verify=True):
        """Read VDIF Header from file.

        Parameters
        ----------
        fh : filehandle
            To read data from.
        edv : int, False, or None, optional
            Extended data version.  If `False`, a legacy header is used.
            If `None` (default), it is determined from the header.  (Given it
            explicitly is mostly useful for a slight speed-up.)
        verify : bool, optional
            Whether to do basic verification of integrity.  Default: `True`.
        """
        # Assume non-legacy header to ensure those are done fastest.
        # Since a payload will follow, it is OK to read too many bytes even
        # for a legacy header; we just rewind below.
        s = fh.read(32)
        if len(s) != 32:
            raise EOFError
        self = cls(eight_word_struct.unpack(s), edv, verify=False)
        if self.edv is False:
            # Legacy headers are 4 words, so rewind, and remove excess data.
            fh.seek(-16, 1)
            self.words = self.words[:4]
        if verify:
            self.verify()

        return self

    @classmethod
    def fromvalues(cls, edv=False, **kwargs):
        """Initialise a header from parsed values.

        Here, the parsed values must be given as keyword arguments, i.e., for
        any ``header = cls(<data>)``, ``cls.fromvalues(**header) == header``.

        However, unlike for the :meth:`~baseband.vdif.VDIFHeader.fromkeys`
        class method, data can also be set using arguments named after methods,
        such as ``bps`` and ``time``.

        Given defaults:

        invalid_data : `False`
        legacy_mode : `False`
        vdif_version : 1
        thread_id : 0
        frame_nr : 0
        sync_pattern : 0xACABFEED for EDV 1 and 3, 0xa5ea5 for EDV 2

        Values set by other keyword arguments (if present):

        bits_per_sample : from ``bps``
        frame_length : from ``samples_per_frame`` or ``frame_nbytes``
        lg2_nchan : from ``nchan``
        station_id : from ``station``
        sampling_rate, sampling_unit : from ``sample_rate``
        ref_epoch, seconds, frame_nr : from ``time``

        Note that to set ``time`` to non-integer seconds one also needs to
        pass in ``frame_rate`` or ``sample_rate``.
        """
        # TODO: here, we act differently depending on what information the
        # VDIF header subclass can handle. Ideally, these different actions
        # would be stored on the subclasses themselves, but then their super
        # calls would end up here anyway.  One could by-pass those by checking
        # whether cls is VDIFHeader.  But another issue is that the headers
        # with a sample rate are currently subclasses of those without, so
        # the super chain doesn't work well.
        #
        # Some defaults that are not done by setting properties.
        kwargs.setdefault('legacy_mode', True if edv is False else False)
        kwargs['edv'] = edv
        # For setting time, one normally needs a frame rate.  If headers
        # provide this information, we can just proceed.
        if 'time' not in kwargs or 'frame_rate' in VDIF_HEADER_CLASSES.get(
                edv if edv is not False else -1, VDIFBaseHeader)._properties:
            return super(VDIFHeader, cls).fromvalues(edv, **kwargs)
        # If the VDIF header subclass does not provide the frame rate, we
        # first initialize without time, and then set the time explicitly
        # using whatever frame_rate or sample_rate was passed.
        time = kwargs.pop('time')
        sample_rate = kwargs.pop('sample_rate', None)
        frame_rate = kwargs.pop('frame_rate', None)
        # Pop verify and pass on False so verify happens after time is set.
        verify = kwargs.pop('verify', True)
        self = super(VDIFHeader, cls).fromvalues(edv, verify=False, **kwargs)
        if frame_rate is None and sample_rate is not None:
            frame_rate = sample_rate / self.samples_per_frame
        self.set_time(time, frame_rate=frame_rate)
        if verify:
            self.verify()
        return self

    @classmethod
    def fromkeys(cls, **kwargs):
        """Initialise a header from parsed values.

        Like :meth:`~baseband.vdif.VDIFHeader.fromvalues`, but without any
        interpretation of keywords.

        Raises
        ------
        KeyError : if not all keys required are pass in.
        """
        # Get all required values.
        edv = False if kwargs['legacy_mode'] else kwargs['edv']
        return super(VDIFHeader, cls).fromkeys(edv, **kwargs)

    @classmethod
    def from_mark5b_header(cls, mark5b_header, bps, nchan, **kwargs):
        """Construct an Mark5B over VDIF header (EDV=0xab).

        See http://www.vlbi.org/vdif/docs/vdif_extension_0xab.pdf

        Note that the Mark 5B header does not encode the bits-per-sample and
        the number of channels used in the payload, so these need to be given
        separately.  A complete frame can be encapsulated with
        `~baseband.vdif.VDIFFrame.from_mark5b_frame`.

        Parameters
        ----------
        mark5b_header : `~baseband.mark5b.Mark5BHeader`
            Used to set time, etc.
        bps : int
            Bits per elementary sample.
        nchan : int
            Number of channels carried in the Mark 5B payload.
        **kwargs
            Any further arguments.  Strictly, none are necessary to create a
            valid VDIF header, but this can be used to pass on, e.g.,
            ``invalid_data``.
        """
        assert 'time' not in kwargs, "Time is inferred from Mark 5B Header."
        # We need to treat the time carefully since the Mark 5B time may miss
        # fractional seconds, or they may be incorrect for rates above 512Mbps.
        # Integer seconds and frame_nr define everything uniquely, so we get
        # the time for frame number 0 and then set the frame number after.
        kwargs.update(mark5b_header)
        # Get time to integer seconds from mark5b header.
        time_frame0 = Time(mark5b_header.kday + mark5b_header.jday,
                           mark5b_header.seconds / 86400,
                           format='mjd', scale='utc')
        self = cls.fromvalues(edv=0xab, bps=bps, nchan=nchan,
                              complex_data=False, time=time_frame0, **kwargs)
        # Time setting will set frame_nr and bcd_fraction to 0; reset these
        # to the correct values.
        self['frame_nr'] = mark5b_header['frame_nr']
        self['bcd_fraction'] = mark5b_header['bcd_fraction']
        return self

    # Properties common to all VDIF headers.
    @property
    def edv(self):
        """VDIF Extended Data Version (EDV)."""
        return self._edv

    @property
    def frame_nbytes(self):
        """Size of the frame in bytes."""
        return self['frame_length'] * 8

    @frame_nbytes.setter
    def frame_nbytes(self, nbytes):
        assert nbytes % 8 == 0
        self['frame_length'] = int(nbytes) // 8

    @property
    def payload_nbytes(self):
        """Size of the payload in bytes."""
        return self.frame_nbytes - self.nbytes

    @payload_nbytes.setter
    def payload_nbytes(self, nbytes):
        self.frame_nbytes = nbytes + self.nbytes

    @property
    def bps(self):
        """Bits per elementary sample."""
        return self['bits_per_sample'] + 1

    @bps.setter
    def bps(self, bps):
        assert bps % 1 == 0
        self['bits_per_sample'] = int(bps) - 1

    @property
    def nchan(self):
        """Number of channels in the frame."""
        return 2**self['lg2_nchan']

    @nchan.setter
    def nchan(self, nchan):
        lg2_nchan = np.log2(nchan)
        assert lg2_nchan % 1 == 0
        self['lg2_nchan'] = int(lg2_nchan)

    @property
    def samples_per_frame(self):
        """Number of complete samples in the frame."""
        # Values are not split over word boundaries.
        values_per_word = 32 // self.bps // (2 if self['complex_data'] else 1)
        # samples are not split over payload boundaries.
        return self.payload_nbytes // 4 * values_per_word // self.nchan

    @samples_per_frame.setter
    def samples_per_frame(self, samples_per_frame):
        values_per_word = 32 // self.bps // (2 if self['complex_data'] else 1)
        # units of frame length are 8 bytes, i.e., 2 words.
        values_per_long = values_per_word * 2
        longs = (samples_per_frame * self.nchan - 1) // values_per_long + 1
        self['frame_length'] = int(longs) + self.nbytes // 8

    @property
    def station(self):
        """Station ID: two ASCII characters, or 16-bit int."""
        msb = self['station_id'] >> 8
        if 48 <= msb < 128:
            return chr(msb) + chr(self['station_id'] & 0xff)
        else:
            return self['station_id']

    @station.setter
    def station(self, station):
        try:
            station_id = (ord(station[0]) << 8) + ord(station[1])
        except TypeError:
            station_id = station
        assert int(station_id) == station_id
        self['station_id'] = station_id

    def get_time(self, frame_rate=None):
        """Converts ref_epoch, seconds, and frame_nr to Time object.

        Uses 'ref_epoch', which stores the number of half-years from 2000,
        and 'seconds'.  By default, it also calculates the offset using
        the current frame number.  For non-zero 'frame_nr', this requires the
        frame rate, which is calculated from the sample rate in the header.

        Parameters
        ----------
        frame_rate : `~astropy.units.Quantity`, optional
            For non-zero 'frame_nr', this is required to calculate the
            corresponding offset.

        Returns
        -------
        time : `~astropy.time.Time`
        """
        frame_nr = self['frame_nr']
        if frame_nr == 0:
            offset = 0.
        else:
            if frame_rate is None:
                raise ValueError("this header does not provide a frame "
                                 "rate. Pass it in explicitly.")

            offset = (frame_nr / frame_rate).to_value(u.s)

        return (ref_epochs[self['ref_epoch']] +
                TimeDelta(self['seconds'], offset, format='sec', scale='tai'))

    def set_time(self, time, frame_rate=None):
        """Converts Time object to ref_epoch, seconds, and frame_nr.

        For non-integer seconds, a frame rate is needed to calculate the
        'frame_nr'.

        Parameters
        ----------
        time : `~astropy.time.Time`
            The time to use for this header.
        frame_rate : `~astropy.units.Quantity`, optional
            For calculating 'frame_nr' from the fractional seconds.
        """
        assert time > ref_epochs[0]
        ref_index = np.searchsorted((ref_epochs - time).sec, 0) - 1
        self['ref_epoch'] = ref_index
        seconds = time - ref_epochs[ref_index]
        int_sec = int(seconds.sec)

        # Round to nearest ns to handle timestamp difference errors.
        frac_sec = seconds - int_sec * u.s
        if abs(frac_sec) < 1. * u.ns:
            frame_nr = 0
        elif abs(1. * u.s - frac_sec) < 1. * u.ns:
            int_sec += 1
            frame_nr = 0
        else:
            if frame_rate is None:
                raise ValueError("this header does not provide a frame "
                                 "rate. Pass it in explicitly.")

            frame_nr = int(round((frac_sec * frame_rate).to_value(u.one)))
            if abs(frame_nr / frame_rate - 1. * u.s) < 1. * u.ns:
                frame_nr = 0
                int_sec += 1

        self['seconds'] = int_sec
        self['frame_nr'] = frame_nr

    time = property(get_time, set_time)


class VDIFLegacyHeader(VDIFHeader):
    """Legacy VDIF header that uses only 4 32-bit words.

    See Section 6 of
    http://www.vlbi.org/vdif/docs/VDIF_specification_Release_1.1.1.pdf
    """
    _struct = four_word_struct

    _header_parser = HeaderParser(
        (('invalid_data', (0, 31, 1, False)),
         ('legacy_mode', (0, 30, 1, True)),
         ('seconds', (0, 0, 30)),
         ('_1_30_2', (1, 30, 2, 0x0)),
         ('ref_epoch', (1, 24, 6)),
         ('frame_nr', (1, 0, 24, 0x0)),
         ('vdif_version', (2, 29, 3, 0x1)),
         ('lg2_nchan', (2, 24, 5)),
         ('frame_length', (2, 0, 24)),
         ('complex_data', (3, 31, 1)),
         ('bits_per_sample', (3, 26, 5)),
         ('thread_id', (3, 16, 10, 0x0)),
         ('station_id', (3, 0, 16))))
    # Set default
    _edv = False

    def verify(self):
        """Basic checks of header integrity."""
        assert self.edv is False
        assert self['legacy_mode']
        assert len(self.words) == 4


class VDIFBaseHeader(VDIFHeader):
    """Base for non-legacy VDIF headers that use 8 32-bit words."""

    _header_parser = VDIFLegacyHeader._header_parser + HeaderParser(
        (('legacy_mode', (0, 30, 1, False)),  # Repeat, to change default.
         ('edv', (4, 24, 8))))

    def verify(self):
        """Basic checks of header integrity."""
        assert not self['legacy_mode']
        assert self.edv is None or self.edv == self['edv']
        assert len(self.words) == 8
        # _sync_pattern is added by VDIFHeaderMeta.
        if 'sync_pattern' in self.keys():
            assert self['sync_pattern'] == self._sync_pattern


class VDIFHeader0(VDIFBaseHeader):
    """VDIF Header for EDV=0.

    EDV=0 implies the extended user data fields are not used.
    """
    _edv = 0

    def verify(self):
        assert all(word == 0 for word in self.words[4:])
        super(VDIFHeader0, self).verify()


class VDIFSampleRateHeader(VDIFBaseHeader):
    """Base for VDIF headers that include the sample rate (EDV= 1, 3, 4)."""

    _header_parser = VDIFBaseHeader._header_parser + HeaderParser(
        (('sampling_unit', (4, 23, 1)),
         ('sampling_rate', (4, 0, 23)),
         ('sync_pattern', (5, 0, 32, 0xACABFEED))))

    # Add extra properties, ensuring 'time' comes after 'sample_rate' and
    # 'frame_rate', since time setting requires the frame rate.
    _properties = (VDIFBaseHeader._properties[:-1] +
                   ('sample_rate', 'frame_rate', 'time'))

    def same_stream(self, other):
        return (super(VDIFSampleRateHeader, self).same_stream(other) and
                self.words[4] == other.words[4] and
                self.words[5] == other.words[5])

    @property
    def sample_rate(self):
        """Number of complete samples per second.

        Assumes the 'sampling_rate' header field represents a per-channel
        sample rate for complex samples, or half the sample rate for real ones.
        """
        # Interprets sample rate correctly for EDV=3, but may not for EDV=1.
        return u.Quantity(self['sampling_rate'] *
                          (1 if self['complex_data'] else 2),
                          u.MHz if self['sampling_unit'] else u.kHz)

    @sample_rate.setter
    def sample_rate(self, sample_rate):
        # Sets sample rate correctly for EDV=3, but may not for EDV=1 or for
        # multiple channels per frame (illegal for EDV=3).
        assert sample_rate.to_value(u.Hz) % 1 == 0
        complex_sample_rate = sample_rate / (1 if self['complex_data'] else 2)
        self['sampling_unit'] = not (
            complex_sample_rate.unit == u.kHz or
            complex_sample_rate.to_value(u.MHz) % 1 != 0)
        if self['sampling_unit']:
            self['sampling_rate'] = int(complex_sample_rate.to_value(u.MHz))
        else:
            assert complex_sample_rate.to(u.kHz).value % 1 == 0
            self['sampling_rate'] = int(complex_sample_rate.to_value(u.kHz))

    @property
    def frame_rate(self):
        """Number of frames per second.

        Assumes the 'sampling_rate' header field represents a per-channel
        sample rate for complex samples, or half the sample rate for real ones.
        """
        return self.sample_rate / self.samples_per_frame

    @frame_rate.setter
    def frame_rate(self, frame_rate):
        self.sample_rate = frame_rate * self.samples_per_frame

    def get_time(self, frame_rate=None):
        """Converts ref_epoch, seconds, and frame_nr to Time object.

        Uses 'ref_epoch', which stores the number of half-years from 2000,
        and 'seconds'.  By default, it also calculates the offset using
        the current frame number.  For non-zero 'frame_nr', this requires the
        frame rate, which is calculated from the sample rate in the header.
        The latter can also be explicitly passed on.

        Parameters
        ----------
        frame_rate : `~astropy.units.Quantity`, optional
            For non-zero 'frame_nr', this is used to calculate the
            corresponding offset. If not given, the frame rate from the
            header is used (if it is non-zero).

        Returns
        -------
        time : `~astropy.time.Time`
        """
        if frame_rate is None and self['sampling_rate'] != 0:
            frame_rate = self.frame_rate
        return super(VDIFSampleRateHeader,
                     self).get_time(frame_rate=frame_rate)

    def set_time(self, time, frame_rate=None):
        """Converts Time object to ref_epoch, seconds, and frame_nr.

        Parameters
        ----------
        time : `~astropy.time.Time`
            The time to use for this header.
        frame_rate : `~astropy.units.Quantity`, optional
            For calculating 'frame_nr' from the fractional seconds. If not
            given, the frame rate from the header is used (if it is non-zero).
        """
        if frame_rate is None and self['sampling_rate'] != 0:
            frame_rate = self.frame_rate
        super(VDIFSampleRateHeader, self).set_time(time, frame_rate=frame_rate)

    time = property(get_time, set_time)


class VDIFHeader1(VDIFSampleRateHeader):
    """VDIF Header for EDV=1.

    See http://www.vlbi.org/vdif/docs/vdif_extension_0x01.pdf
    """
    _edv = 1
    _header_parser = VDIFSampleRateHeader._header_parser + HeaderParser(
        (('das_id', (6, 0, 64, 0x0)),))


class VDIFHeader3(VDIFSampleRateHeader):
    """VDIF Header for EDV=3.

    See http://www.vlbi.org/vdif/docs/vdif_extension_0x03.pdf
    """
    _edv = 3
    _header_parser = VDIFSampleRateHeader._header_parser + HeaderParser(
        (('frame_length', (2, 0, 24, 629)),  # Repeat, to set default.
         ('loif_tuning', (6, 0, 32, 0x0)),
         ('_7_28_4', (7, 28, 4, 0x0)),
         ('dbe_unit', (7, 24, 4, 0x0)),
         ('if_nr', (7, 20, 4, 0x0)),
         ('subband', (7, 17, 3, 0x0)),
         ('sideband', (7, 16, 1, False)),
         ('major_rev', (7, 12, 4, 0x0)),
         ('minor_rev', (7, 8, 4, 0x0)),
         ('personality', (7, 0, 8))))

    def verify(self):
        super(VDIFHeader3, self).verify()
        assert self['frame_length'] == 629


class VDIFHeader2(VDIFBaseHeader):
    """VDIF Header for EDV=2.

    See http://www.vlbi.org/vdif/docs/alma-vdif-edv.pdf

    Notes
    -----
    This header is untested.  It may need to have subclasses, based on possible
    different sync values.
    """
    _edv = 2

    _header_parser = VDIFBaseHeader._header_parser + HeaderParser(
        (('complex_data', (3, 31, 1, 0x0)),  # Repeat, to set default.
         ('bits_per_sample', (3, 26, 5, 0x1)),  # Repeat, to set default.
         ('pol', (4, 0, 1)),
         ('BL_quadrant', (4, 1, 2)),
         ('BL_correlator', (4, 3, 1)),
         ('sync_pattern', (4, 4, 20, 0xa5ea5)),
         ('PIC_status', (5, 0, 32)),
         ('PSN', (6, 0, 64))))

    def verify(self):  # pragma: no cover
        super(VDIFHeader2, self).verify()
        assert self['frame_length'] == 629 or self['frame_length'] == 1004
        assert self.bps == 2 and not self['complex_data']


class VDIFMark5BHeader(VDIFBaseHeader, Mark5BHeader):
    """Mark 5B over VDIF (EDV=0xab).

    See http://www.vlbi.org/vdif/docs/vdif_extension_0xab.pdf
    """
    _edv = 0xab
    # Repeat 'frame_length' to set default.
    _header_parser = (VDIFBaseHeader._header_parser +
                      HeaderParser((('frame_length', (2, 0, 24, 1254)),)) +
                      HeaderParser(tuple(
                          ((k if k != 'frame_nr' else 'mark5b_frame_nr'),
                           (v[0] + 4,) + v[1:])
                          for (k, v) in Mark5BHeader._header_parser.items())))

    def verify(self):
        super(VDIFMark5BHeader, self).verify()
        assert self['frame_length'] == 1254  # payload+header=10000+32 bytes/8
        assert self['frame_nr'] == self['mark5b_frame_nr']
        # Check consistency of time down to the second (since some Mark 5B
        # headers do not store 'bcd_fraction').
        day, seconds = divmod(self['seconds'], 86400)
        assert seconds == self.seconds  # Latter decodes 'bcd_seconds'
        ref_mjd = ref_epochs[self['ref_epoch']].mjd + day
        assert ref_mjd % 1000 == self.jday  # Latter decodes 'bcd_jday'

    def __setitem__(self, item, value):
        super(VDIFMark5BHeader, self).__setitem__(item, value)
        if item == 'frame_nr':
            super(VDIFMark5BHeader, self).__setitem__('mark5b_frame_nr', value)

    def get_time(self, frame_rate=None):
        """
        Convert ref_epoch, seconds, and fractional seconds to Time object.

        Uses 'ref_epoch', which stores the number of half-years from 2000,
        and 'seconds', from the VDIF part of the header, and the fractional
        seconds from the Mark 5B part.

        Since some Mark 5B headers do not store the fractional seconds,
        one can also calculates the offset using the current frame number by
        passing in a sample rate.

        Furthermore, fractional seconds are stored only to 0.1 ms accuracy.
        In the code, this is "unrounded" to give the exact time of the start
        of the frame for any total bit rate below 512 Mbps.  For rates above
        this value, it is no longer guaranteed that subsequent frames have
        unique rates, and one should pass in an explicit sample rate instead.

        Parameters
        ----------
        frame_rate : `~astropy.units.Quantity`, optional
            For non-zero 'frame_nr', this is used to calculate the
            corresponding offset.

        Returns
        -------
        time : `~astropy.time.Time`
        """
        frame_nr = self['frame_nr']
        if frame_nr == 0:
            fraction = 0.
        elif frame_rate is None:
            # Get fractional second from the Mark 5B part of the header,
            # but check it is non-zero (it doesn't always seem to be set).
            fraction = self.fraction
            if fraction == 0.:
                raise ValueError('header does not provide correct fractional '
                                 'second (it is zero for non-zero frame '
                                 'number). Please pass in a frame_rate.')
        else:
            fraction = (frame_nr / frame_rate).to_value(u.s)

        return (ref_epochs[self['ref_epoch']] +
                TimeDelta(self['seconds'], fraction,
                          format='sec', scale='tai'))

    def set_time(self, time, frame_rate=None):
        Mark5BHeader.set_time(self, time, frame_rate)
        super(VDIFMark5BHeader, self).set_time(time, frame_rate)

    time = property(get_time, set_time)



# Licensed under the GPLv3 - see LICENSE
"""Radio baseband I/O."""

# Affiliated packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *
# ----------------------------------------------------------------------------

# For egg_info test builds to pass, put package imports here.
if not _ASTROPY_SETUP_:
    # from example_mod import *
    pass


from .core import file_info, open



# Licensed under the GPLv3 - see LICENSE
"""Payload for GUPPI format."""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import numpy as np
from collections import namedtuple

from ..vlbi_base.payload import VLBIPayloadBase


__all__ = ['GUPPIPayload']


def decode_8bit(words):
    return words.view(np.int8, np.ndarray).astype(np.float32)


def encode_8bit(values):
    return np.clip(np.rint(values), -128, 127).astype(np.int8)


class GUPPIPayload(VLBIPayloadBase):
    """Container for decoding and encoding GUPPI payloads.

    Parameters
    ----------
    words : `~numpy.ndarray`
        Array containg LSB unsigned words (with the right size) that
        encode the payload.
    header : `~baseband.guppi.GUPPIHeader`
        Header that provides information about how the payload is encoded.
        If not given, the following arguments have to be passed in.
    bps : int, optional
        Number of bits per sample part (i.e., per channel and per real or
        imaginary component).  Default: 8.
    sample_shape : tuple, optional
        Shape of the samples; e.g., (nchan,).  Default: ().
    complex_data : bool, optional
        Whether data are complex.  Default: `False`.
    channels_first : bool, optional
        Whether the encoded payload is stored as (nchan, nsample, npol),
        rather than (nsample, nchan, npol).  Default: `True`.
    """
    _decoders = {
        8: decode_8bit}
    _encoders = {
        8: encode_8bit}
    _dtype_word = np.dtype('int8')

    _sample_shape_maker = namedtuple('SampleShape', 'npol, nchan')

    def __init__(self, words, header=None, sample_shape=(), bps=8,
                 complex_data=False, channels_first=True):
        if header is not None:
            bps = header.bps
            sample_shape = header.sample_shape
            complex_data = header.complex_data
            channels_first = header.channels_first
        super(GUPPIPayload, self).__init__(words, sample_shape=sample_shape,
                                           bps=bps, complex_data=complex_data)
        self.channels_first = channels_first
        # If channels first, _item_to_slices must act on per-channel words.  By
        # resetting self._bpfs, we allow _item_to_slices to work unmodified.
        self._true_bpfs = self._bpfs    # Save the true bpfs regardless.
        if self.channels_first:
            self._bpfs //= self.sample_shape.nchan

    @classmethod
    def fromfile(cls, fh, header=None, memmap=False, payload_nbytes=None,
                 **kwargs):
        """Read or map encoded data in file.

        Parameters
        ----------
        fh : filehandle
            Handle to the file which will be read or mapped.
        header : `~baseband.guppi.GUPPIHeader`, optional
            If given, used to infer ``payload_nbytes``, ``bps``,
            ``sample_shape``, ``complex_data`` and ``channels_first``.  If not
            given, those have to be passed in.
        memmap : bool, optional
            If `False` (default), read from file.  Otherwise, map the file in
            memory (see `~numpy.memmap`).
        payload_nbytes : int, optional
            Number of bytes to read (default: as given in ``header``,
            ``cls._nbytes``, or, for mapping, to the end of the file).
        **kwargs
            Additional arguments are passed on to the class initializer. These
            are only needed if ``header`` is not given.
        """
        if payload_nbytes is None:
            payload_nbytes = (cls._nbytes if header is None
                              else header.payload_nbytes)

        if not memmap:
            return super(GUPPIPayload, cls).fromfile(
                fh, header=header, payload_nbytes=payload_nbytes, **kwargs)

        words_shape = (payload_nbytes // cls._dtype_word.itemsize,)
        if hasattr(fh, 'memmap'):
            words = fh.memmap(dtype=cls._dtype_word,
                              shape=(None if payload_nbytes is None else
                                     words_shape))
        else:
            mode = fh.mode.replace('b', '')
            offset = fh.tell()
            words = np.memmap(
                fh, mode=mode, dtype=cls._dtype_word, offset=offset,
                shape=(None if payload_nbytes is None else words_shape))
            fh.seek(offset + words.size * words.dtype.itemsize)
        return cls(words, header=header, **kwargs)

    @classmethod
    def fromdata(cls, data, header=None, bps=8, channels_first=True):
        """Encode data as a payload.

        Parameters
        ----------
        data : `~numpy.ndarray`
            Data to be encoded. The last dimension is taken as the number of
            channels.
        header : `~baseband.guppi.GUPPIHeader`, optional
            If given, used to infer the ``bps`` and ``channels_first``.
        bps : int, optional
            Bits per elementary sample, used if ``header`` is `None`.
            Default: 8.
        channels_first : bool, optional
            Whether encoded data should be ordered as (nchan, nsample, npol),
            used if ``header`` is `None`.  Default: `True`.
        """
        if header is not None:
            bps = header.bps
            channels_first = header.channels_first
        sample_shape = data.shape[1:]
        complex_data = data.dtype.kind == 'c'
        try:
            encoder = cls._encoders[bps]
        except KeyError:
            raise ValueError("{0} cannot encode data with {1} bits"
                             .format(cls.__name__, bps))
        # If channels-first, switch to (nchan, nsample, npol); otherwise use
        # (nsample, nchan, npol).
        if channels_first:
            data = data.transpose(2, 0, 1)
        else:
            data = data.transpose(0, 2, 1)
        if complex_data:
            data = data.view((data.real.dtype, (2,)))
        words = encoder(data).ravel().view(cls._dtype_word)
        return cls(words, sample_shape=sample_shape, bps=bps,
                   complex_data=complex_data, channels_first=channels_first)

    def __len__(self):
        """Number of samples in the payload."""
        return self.nbytes * 8 // self._true_bpfs

    def __getitem__(self, item=()):
        # GUPPI data may be stored as (nsample, nchan, npol) or, if
        # channels-first, (nchan, nsample, npol), both of which require
        # reshaping to get the usual order of (nsample, npol, nchan).
        decoder = self._decoders[self._coder]

        # If we want to decode the entire dataset.
        if item is () or item == slice(None):
            data = decoder(self.words)
            if self.complex_data:
                data = data.view(self.dtype)
            if self.channels_first:
                # Reshape to (nchan, nsample, npol); transpose to usual order.
                return (data.reshape(self.sample_shape.nchan, -1,
                                     self.sample_shape.npol)
                        .transpose(1, 2, 0))
            else:
                # Reshape to (nsample, nchan, npol); transpose to usual order.
                return (data.reshape(-1, self.sample_shape.nchan,
                                     self.sample_shape.npol)
                        .transpose(0, 2, 1))

        words_slice, data_slice = self._item_to_slices(item)

        if self.channels_first:
            # Reshape words so channels fall along first axis, then decode.
            decoded_words = decoder(
                self.words.reshape(self.sample_shape.nchan, -1)[:, words_slice])
            # Reshape to (nsample, nchan, npol), then use data_slice.
            return (decoded_words.view(self.dtype).T
                    .reshape(-1, *self.sample_shape)[data_slice])
        else:
            # data_slice assumes (npol, nchan), so transpose before using it.
            return (decoder(self.words[words_slice]).view(self.dtype)
                    .reshape(-1, self.sample_shape.nchan,
                             self.sample_shape.npol)
                    .transpose(0, 2, 1)[data_slice])

    def __setitem__(self, item, data):
        if item is () or item == slice(None):
            words_slice = data_slice = slice(None)
        else:
            words_slice, data_slice = self._item_to_slices(item)

        data = np.asanyarray(data)
        # Check if the new data spans an entire word and is correctly shaped.
        # If so, skip decoding.  If not, decode appropriate words and insert
        # new data.
        if not (data_slice == slice(None) and
                data.shape[-2:] == self.sample_shape and
                data.dtype.kind == self.dtype.kind):
            decoder = self._decoders[self._coder]
            if self.channels_first:
                decoded_words = decoder(np.ascontiguousarray(
                    self.words.reshape(
                        self.sample_shape.nchan, -1)[:, words_slice]))
                current_data = (decoded_words.view(self.dtype)
                                .T.reshape(-1, *self.sample_shape))
            else:
                current_data = (decoder(self.words[words_slice])
                                .view(self.dtype)
                                .reshape(-1, self.sample_shape.nchan,
                                         self.sample_shape.npol)
                                .transpose(0, 2, 1))

            current_data[data_slice] = data
            data = current_data

        # Reshape before separating real and complex components.
        if self.channels_first:
            data = data.reshape(-1, self.sample_shape.nchan).T
        else:
            data = data.transpose(0, 2, 1)

        # Separate real and complex components.
        if data.dtype.kind == 'c':
            data = data.view((data.real.dtype, (2,)))

        # Select encoder.
        encoder = self._encoders[self._coder]

        # Reshape and encode words.
        if self.channels_first:
            self.words.reshape(self.sample_shape.nchan, -1)[:, words_slice] = (
                encoder(data).reshape(self.sample_shape.nchan, -1)
                .view(self._dtype_word))
        else:
            self.words[words_slice] = (encoder(data.ravel())
                                       .view(self._dtype_word))

    data = property(__getitem__, doc="Full decoded payload.")



# Licensed under the GPLv3 - see LICENSE
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from ..vlbi_base.file_info import VLBIFileReaderInfo


class Mark5BFileReaderInfo(VLBIFileReaderInfo):
    _header0_attrs = ()
    _parent_attrs = ('nchan', 'bps', 'kday', 'ref_time')

    def _get_start_time(self):
        try:
            return self.header0.get_time(frame_rate=self.frame_rate)
        except Exception:
            return None

    def _collect_info(self):
        super(Mark5BFileReaderInfo, self)._collect_info()
        if self:
            self.complex_data = False
            if self.kday is None and self.ref_time is None:
                self.missing['kday'] = self.missing['ref_time'] = (
                    "needed to infer full times.")

            if self.nchan is None:
                self.missing['nchan'] = (
                    "needed to determine sample shape and rate.")
            else:
                self.sample_shape = (self.nchan,)
                self.samples_per_frame = (self.header0.payload_nbytes * 8 //
                                          (self.bps * self.nchan))
                self.sample_rate = self.samples_per_frame * self.frame_rate



# Licensed under the GPLv3 - see LICENSE
"""GMRT Software Backend (GSB) data reader.

See http://gmrt.ncra.tifr.res.in/gmrt_hpage/sub_system/gmrt_gsb/index.htm
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from .base import open
from .header import GSBHeader
from .payload import GSBPayload
from .frame import GSBFrame
